from __future__ import annotations

import json
from pathlib import Path

import traitlets as tl
from importlib_resources import files
from jinja2 import Environment

from aiida import orm
from aiida.cmdline.utils.common import get_workchain_report
from aiida_quantumespresso.workflows.pw.bands import PwBandsWorkChain
from aiidalab_qe.app.parameters import DEFAULT_PARAMETERS
from aiidalab_qe.app.result.components import ResultsComponentModel
from aiidalab_qe.app.static import styles, templates
from aiidalab_qe.common.time import format_time, relative_time

DEFAULT: dict = DEFAULT_PARAMETERS  # type: ignore


FUNCTIONAL_LINK_MAP = {
    "PBE": "https://journals.aps.org/prl/abstract/10.1103/PhysRevLett.77.3865",
    "PBEsol": "https://journals.aps.org/prl/abstract/10.1103/PhysRevLett.100.136406",
}

PSEUDO_LINK_MAP = {
    "SSSP": "https://www.materialscloud.org/discover/sssp/table/efficiency",
    "PseudoDojo": "http://www.pseudo-dojo.org/",
}

FUNCTIONAL_REPORT_MAP = {
    "LDA": "local density approximation (LDA)",
    "PBE": "generalized gradient approximation of Perdew-Burke-Ernzerhof (PBE)",
    "PBEsol": "the revised generalized gradient approximation of Perdew-Burke-Ernzerhof (PBE) for solids",
}

# Periodicity
PERIODICITY_MAPPING = {
    (True, True, True): "xyz",
    (True, True, False): "xy",
    (True, False, False): "x",
    (False, False, False): "molecule",
}

VDW_CORRECTION_VERSION = {
    3: "Grimme-D3",
    4: "Grimme-D3BJ",
    5: "Grimme-D3M",
    6: "Grimme-D3MBJ",
    "ts-vdw": "Tkatchenko-Scheffler",
    "none": "None",
}


class WorkChainSummaryModel(ResultsComponentModel):
    identifier = "workflow summary"

    failed_calculation_report = tl.Unicode("")

    has_failure_report = False

    @property
    def include(self):
        return True

    def generate_report_html(self):
        """Read from the builder parameters and generate a html for reporting
        the inputs for the `QeAppWorkChain`.
        """
        env = Environment()
        template = files(templates).joinpath("workflow_summary.jinja").read_text()
        parameters = self._generate_report_parameters()
        report = {key: value for key, value in parameters.items() if value is not None}
        schema = json.load(Path(__file__).parent.joinpath("schema.json").open())
        return env.from_string(template).render(
            report=report,
            schema=schema,
            format=DEFAULT["summary_format"],
        )

    def generate_report_text(self, report_dict):
        """Generate a text for reporting the inputs for the `QeAppWorkChain`

        :param report_dict: dictionary generated by the `generate_report_dict` function.
        """
        report_string = (
            "All calculations are performed within the density-functional "
            "theory formalism as implemented in the Quantum ESPRESSO code. "
            "The pseudopotential for each element is extracted from the "
            f'{report_dict["Pseudopotential library"][0]} '
            "library. The wave functions "
            "of the valence electrons are expanded in a plane wave basis set, using an "
            "energy cutoff equal to "
            f'{round(report_dict["Plane wave energy cutoff (wave functions)"][0])} Ry '
            "for the wave functions and "
            f'{round(report_dict["Plane wave energy cutoff (charge density)"][0])} Ry '
            "for the charge density and potential. "
            "The exchange-correlation energy is "
            "calculated using the "
            f'{FUNCTIONAL_REPORT_MAP[report_dict["Functional"][0]]}. '
            "A Monkhorst-Pack mesh is used for sampling the Brillouin zone, where the "
            "distance between the k-points is set to "
        )
        kpoints_distances = []
        kpoints_calculations = []

        for calc in ("SCF", "NSCF", "Bands"):
            if f"K-point mesh distance ({calc})" in report_dict:
                kpoints_distances.append(
                    str(report_dict[f"K-point mesh distance ({calc})"][0])
                )
                kpoints_calculations.append(calc)

        report_string += ", ".join(kpoints_distances)
        report_string += " for the "
        report_string += ", ".join(kpoints_calculations)
        report_string += " calculation"
        if len(kpoints_distances) > 1:
            report_string += "s, respectively"
        report_string += "."

        return report_string

    def generate_failure_report(self):
        """Generate a html for reporting the failure of the `QeAppWorkChain`."""
        process_node = self.fetch_process_node()
        if not (process_node and process_node.exit_status):
            return
        final_calcjob = self._get_final_calcjob(process_node)
        env = Environment()
        template = files(templates).joinpath("workflow_failure.jinja").read_text()
        style = files(styles).joinpath("style.css").read_text()
        self.failed_calculation_report = env.from_string(template).render(
            style=style,
            process_report=get_workchain_report(process_node, "REPORT"),
            calcjob_exit_message=final_calcjob.exit_message,
        )
        self.has_failure_report = True

    def _generate_report_parameters(self):
        """Generate the report parameters from the ui parameters and workchain's input.

        Parameters extracted from ui parameters, directly from the widgets,
        such as the ``pseudo_family`` and ``relax_type``.

        Parameters extracted from workchain's inputs, such as the ``energy_cutoff_wfc``
        and ``energy_cutoff_rho``.

        Return a dictionary of the parameters.
        """
        from aiida.orm.utils.serialize import deserialize_unsafe

        qeapp_wc = self.fetch_process_node()

        ui_parameters = qeapp_wc.base.extras.get("ui_parameters", {})
        if isinstance(ui_parameters, str):
            ui_parameters = deserialize_unsafe(ui_parameters)
        # Construct the report parameters needed for the report
        # drop support for old ui parameters
        if "workchain" not in ui_parameters:
            return {}

        inputs = qeapp_wc.inputs
        structure: orm.StructureData = inputs.structure
        basic = ui_parameters["workchain"]
        advanced = ui_parameters["advanced"]
        ctime = qeapp_wc.ctime
        mtime = qeapp_wc.mtime

        report = {
            "workflow_properties": {
                "pk": qeapp_wc.pk,
                "uuid": str(qeapp_wc.uuid),
                "label": qeapp_wc.label,
                "description": qeapp_wc.description,
                "creation_time": f"{format_time(ctime)} ({relative_time(ctime)})",
                "modification_time": f"{format_time(mtime)} ({relative_time(mtime)})",
            },
        }

        report |= {
            "initial_structure_properties": {
                "structure_pk": structure.pk,
                "structure_uuid": structure.uuid,
                "formula": structure.get_formula(),
                "num_atoms": len(structure.sites),
            }
        }

        report["initial_structure_properties"] |= {
            **self._get_symmetry_group_info(structure),
            "cell_lengths": "{:.3f} {:.3f} {:.3f}".format(*structure.cell_lengths),
            "cell_angles": "{:.0f} {:.0f} {:.0f}".format(*structure.cell_angles),
        }

        report |= {
            "basic_settings": {
                "relaxed": "off"
                if basic["relax_type"] == "none"
                else basic["relax_type"],
                "protocol": basic["protocol"],
                "spin_type": "off" if basic["spin_type"] == "none" else "on",
                "electronic_type": basic["electronic_type"],
                "periodicity": PERIODICITY_MAPPING.get(structure.pbc, "xyz"),
            },
            "advanced_settings": {},
        }

        pseudo_family = advanced.get("pseudo_family")
        pseudo_family_info = pseudo_family.split("/")
        pseudo_library = pseudo_family_info[0]
        pseudo_version = pseudo_family_info[1]
        functional = pseudo_family_info[2]
        if pseudo_library == "SSSP":
            pseudo_protocol = pseudo_family_info[3]
        elif pseudo_library == "PseudoDojo":
            pseudo_protocol = pseudo_family_info[4]
        report["advanced_settings"] |= {
            "functional": {
                "url": FUNCTIONAL_LINK_MAP[functional],
                "value": functional,
            },
            "pseudo_library": {
                "url": PSEUDO_LINK_MAP[pseudo_library],
                "value": f"{pseudo_library} {pseudo_protocol} v{pseudo_version}",
            },
        }

        # Extract the pw calculation parameters from the ui_parameters
        pw_parameters = ui_parameters["advanced"].get("pw", {}).get("parameters", {})
        system = pw_parameters["SYSTEM"]
        occupation = "fixed" if basic["electronic_type"] == "insulator" else "smearing"
        report["advanced_settings"] |= {
            "energy_cutoff_wfc": f"{system['ecutwfc']} Ry",
            "energy_cutoff_rho": f"{system['ecutrho']} Ry",
            "occupation_type": occupation,
        }
        if occupation == "smearing":
            report["advanced_settings"] |= {
                "smearing": system["smearing"],
                "degauss": f"{system['degauss']} Ry",
            }
        inv_ang = "Å<sup>-1</sup>"
        kpoints = ui_parameters["advanced"]["kpoints_distance"]
        report["advanced_settings"]["scf_kpoints_distance"] = f"{kpoints} {inv_ang}"
        if "bands" in inputs:
            key = "bands_kpoints_distance"
            kpoints = PwBandsWorkChain.get_protocol_inputs(basic["protocol"])[key]
            report["advanced_settings"][key] = f"{kpoints} {inv_ang}"
        if "pdos" in inputs:
            key = "nscf_kpoints_distance"
            kpoints = inputs.pdos.nscf.kpoints_distance.base.attributes.get("value")
            report["advanced_settings"][key] = f"{kpoints} {inv_ang}"

        vdw_corr = VDW_CORRECTION_VERSION.get(
            system.get("dftd3_version"),
            system.get("vdw_corr", "none"),
        )
        report["advanced_settings"] |= {
            "tot_charge": system.get("tot_charge", 0.0),
            "vdw_corr": "off" if vdw_corr == "none" else vdw_corr,
        }

        if basic["spin_type"] == "collinear":
            if tot_magnetization := system.get("tot_magnetization", False):
                report["advanced_settings"]["tot_magnetization"] = tot_magnetization
            else:
                report["advanced_settings"]["initial_magnetic_moments"] = advanced[
                    "initial_magnetic_moments"
                ]

        if hubbard_dict := ui_parameters["advanced"].pop("hubbard_parameters", None):
            hubbard_parameters = hubbard_dict["hubbard_u"]
            report["advanced_settings"]["hubbard_u"] = hubbard_parameters

        spin_orbit = system.get("lspinorb", False)
        report["advanced_settings"]["spin_orbit"] = "on" if spin_orbit else "off"

        return report

    def _get_symmetry_group_info(self, structure: orm.StructureData) -> dict:
        # HACK the use of the clone for non-molecular systems is due to a rigid
        # condition in AiiDA < 2.6 that only considers 3D systems as pymatgen
        # `Structure` objects (`Molecule` otherwise). Once AiiDAlab is updated with
        # AiiDA 2.6 throughout, we can fall back to using `StructureData.get_pymatgen`
        # (which this method mimics) to obtain the correct pymatgen object
        # (`Molecule` for 0D systems | `Structure` otherwise)
        if any(structure.pbc):
            return self._get_pymatgen_structure(structure)
        return self._get_pymatgen_molecule(structure)

    @staticmethod
    def _get_pymatgen_structure(structure: orm.StructureData) -> dict:
        from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

        clone = structure.clone()
        clone.pbc = (True, True, True)
        analyzer = SpacegroupAnalyzer(structure=clone.get_pymatgen_structure())
        symbol = analyzer.get_space_group_symbol()
        number = analyzer.get_space_group_number()
        return {"space_group": f"{symbol} ({number})"}

    @staticmethod
    def _get_pymatgen_molecule(structure: orm.StructureData) -> dict:
        from pymatgen.symmetry.analyzer import PointGroupAnalyzer

        analyzer = PointGroupAnalyzer(mol=structure.get_pymatgen_molecule())
        return {"point_group": analyzer.get_pointgroup()}

    @staticmethod
    def _get_final_calcjob(node: orm.WorkChainNode) -> orm.CalcJobNode | None:
        """Get the final calculation job node called by a workchain node.

        Parameters
        ----------
        `node`: `orm.WorkChainNode`
            The work chain node to get the final calculation job node from.

        Returns
        -------
        `orm.CalcJobNode` | `None`
            The final calculation job node called by the workchain node if available.
        """
        try:
            final_calcjob = [
                process
                for process in node.called_descendants
                if isinstance(process, orm.CalcJobNode) and process.is_finished
            ][-1]
        except IndexError:
            final_calcjob = None
        return final_calcjob



"""For running the app from the command line used for post_install script."""

import sys
from pathlib import Path

import click

# The default profile name of AiiDAlab container.
_DEFAULT_PROFILE = "default"


@click.group()
def cli():
    pass


@cli.command()
@click.option("-f", "--force", is_flag=True)
@click.option("--computer")
@click.option("-p", "--profile", default=_DEFAULT_PROFILE)
def install_qe(force, profile, computer):
    from aiida import load_profile
    from aiidalab_qe.setup.codes import codes_are_setup, install_and_setup

    load_profile(profile)
    try:
        for msg in install_and_setup(computer=computer, force=force):
            click.echo(msg)
        assert codes_are_setup(computer=computer)
        click.secho("Codes are setup!", fg="green")
    except Exception as error:
        raise click.ClickException(f"Failed to set up QE: {error}") from error


@cli.command()
@click.option("-p", "--profile", default=_DEFAULT_PROFILE, help="AiiDA profile name.")
@click.option(
    "source",
    "--source",
    default=None,
    help="The source folder to install from local.",
    type=click.Path(exists=True, path_type=Path, resolve_path=True),
)
def install_pseudos(profile, source):
    """Install pseudopotentials from a local folder if source is specified,
    otherwise download from remote repositories.
    """
    from aiida import load_profile
    from aiidalab_qe.setup.pseudos import install

    load_profile(profile)

    try:
        for msg, _ in install(cwd=source, download_only=False):
            click.echo(msg)
        click.secho("Pseudopotentials are installed!", fg="green")
    except Exception as error:
        raise click.ClickException(
            f"Failed to set up pseudo potentials: {error}"
        ) from error


@cli.command()
@click.option(
    "dest",
    "--dest",
    default=None,
    help="The dest folder where to download the pseudos.",
    type=click.Path(exists=True, path_type=Path, resolve_path=True),
)
def download_pseudos(dest):
    from aiidalab_qe.setup.pseudos import EXPECTED_PSEUDOS, _install_pseudos

    try:
        for progress in _install_pseudos(
            EXPECTED_PSEUDOS, download_only=True, cwd=dest
        ):
            click.echo(progress)
        click.secho("Pseudopotentials are downloaded!", fg="green")

    except Exception as error:
        raise click.ClickException(
            f"Failed to download pseudo potentials: {error}"
        ) from error


@cli.command()
@click.argument(
    "plugin_name",
    default="aiidalab_qe",
)
@click.option("-p", "--profile", default=_DEFAULT_PROFILE)
def test_plugin(plugin_name, profile):
    from aiida import load_profile
    from aiidalab_qe.app.utils import test_plugin_functionality

    load_profile(profile)

    try:
        success, message = test_plugin_functionality(plugin_name)
        if success:
            click.secho("Plugin is loaded successfully!", fg="green")
        else:
            click.secho(f"Failed to load plugin: {message}", fg="red", err=True)
            sys.exit(1)  # Exit with status 1 to indicate failure
    except Exception as error:
        click.secho(f"Failed to load plugin: {error}", fg="red", err=True)
        sys.exit(1)  # Exit with status 1 to indicate failure


if __name__ == "__main__":
    cli()



"""Widgets for the upload and selection of structure data.

Authors: AiiDAlab team
"""

import pathlib

import ipywidgets as ipw

from aiidalab_qe.app.structure.model import StructureStepModel
from aiidalab_qe.app.utils import get_entry_items
from aiidalab_qe.common import (
    AddingTagsEditor,
    LazyLoadedOptimade,
    LazyLoadedStructureBrowser,
    PeriodicityEditor,
    ShakeNBreakEditor,
)
from aiidalab_qe.common.infobox import InAppGuide
from aiidalab_qe.common.widgets import CategorizedStructureExamplesWidget
from aiidalab_qe.common.wizard import QeConfirmableWizardStep
from aiidalab_widgets_base import (
    BasicCellEditor,
    BasicStructureEditor,
    StructureManagerWidget,
    StructureUploadWidget,
)

# The Examples list of (name, file) tuple currently passed to
# StructureExamplesWidget.
file_path = pathlib.Path(__file__).parent
Examples = [
    ("Bulk silicon", file_path / "examples" / "Si.cif"),
    ("Diamond", file_path / "examples" / "Diamond.cif"),
    ("Gallium arsenide", file_path / "examples" / "GaAs.cif"),
    ("Gold", file_path / "examples" / "Au.cif"),
    ("Nickel", file_path / "examples" / "Ni.cif"),
    ("LiCoO2", file_path / "examples" / "LiCoO2.cif"),
    ("Silicon oxide (alpha quartz)", file_path / "examples" / "SiO2.cif"),
]


class StructureSelectionStep(QeConfirmableWizardStep[StructureStepModel]):
    """Integrated widget for the selection and edition of structure.
    The widget includes a structure manager that allows to select a structure
    from different sources. It also includes the structure editor. Both the
    structure importers and the structure editors can be extended by plugins.
    """

    def __init__(self, model: StructureStepModel, auto_setup=True, **kwargs):
        super().__init__(
            model=model,
            confirm_kwargs={
                "tooltip": "Confirm the currently selected structure and go to the next step",
            },
            **kwargs,
        )
        self._model.observe(
            self._on_installation_change,
            ["installing_sssp", "sssp_installed"],
        )
        self._model.observe(
            self._on_sssp_installed,
            "sssp_installed",
        )
        self._model.observe(
            self._on_input_structure_change,
            "input_structure",
        )
        self._install_sssp(auto_setup)

    def _render(self):
        super()._render()

        examples_by_category = {"Simple crystals": Examples}
        plugin_structure_examples = {
            item["title"]: item["structures"]
            for item in get_entry_items(
                "aiidalab_qe.properties", "structure_examples"
            ).values()
        }
        examples_by_category |= plugin_structure_examples

        importers = [
            StructureUploadWidget(title="Upload file"),
            LazyLoadedOptimade(title="OPTIMADE"),
            LazyLoadedStructureBrowser(title="AiiDA database"),
            CategorizedStructureExamplesWidget(
                title="From examples", examples_by_category=examples_by_category
            ),
        ]

        plugin_importers = get_entry_items("aiidalab_qe.properties", "importer")
        importers.extend([importer() for importer in plugin_importers.values()])

        editors = [
            BasicCellEditor(title="Edit cell"),
            BasicStructureEditor(title="Edit structure"),
            AddingTagsEditor(title="Edit atom tags"),
            PeriodicityEditor(title="Edit periodicity"),
            ShakeNBreakEditor(title="ShakeNBreak"),
        ]

        plugin_editors = get_entry_items("aiidalab_qe.properties", "editor")
        editors.extend([editor() for editor in plugin_editors.values()])

        self.manager = StructureManagerWidget(
            importers=importers,
            editors=editors,
            node_class="StructureData",
            storable=False,
            configuration_tabs=[
                "Cell",
                "Selection",
                "Appearance",
                "Download",
            ],
        )

        if self._model.confirmed:  # loaded from a process
            # NOTE important to do this prior to setting up the links
            # to avoid an override of the structure in the model,
            # which in turn would trigger a reset of the model
            self.manager.input_structure = self._model.input_structure

        ipw.dlink(
            (self.manager, "structure_node"),
            (self._model, "input_structure"),
        )
        ipw.link(
            (self._model, "manager_output"),
            (self.manager.output, "value"),
        )

        self.structure_name_text = ipw.Text(
            placeholder="[No structure selected]",
            description="Selected:",
            disabled=True,
            layout=ipw.Layout(width="auto", flex="1 1 auto"),
        )
        ipw.dlink(
            (self._model, "structure_name"),
            (self.structure_name_text, "value"),
        )

        self.content.children = [
            InAppGuide(identifier="structure-step"),
            ipw.HTML("""
                <p>
                    Select a structure from one of the following sources, then
                    click <span style="color: #4caf50;">
                        <i class="fa fa-check-circle"></i> <b>Confirm</b>
                    </span> to go to the next step
                </p>
            """),
            self.manager,
            self.structure_name_text,
        ]

        self.confirm_box.children += (self.sssp_installation,)

        self.children = [
            self.content,
            self.confirm_box,
        ]

    def _post_render(self):
        # After rendering the widget, nglview needs to be resized
        # to properly display the structure
        self.manager.viewer._viewer.handle_resize()

    def confirm(self, _=None):
        self.manager.store_structure()
        super().confirm()

    def can_reset(self):
        return self._model.confirmed

    def reset(self):
        self._model.reset()

    def _on_installation_change(self, _):
        self._model.update_blockers()

    def _on_sssp_installed(self, _):
        self._toggle_sssp_installation_widget()

    def _on_input_structure_change(self, _):
        self._model.update_widget_text()
        self._update_state()

    def _install_sssp(self, auto_setup):
        from aiidalab_qe.common.setup_pseudos import PseudosInstallWidget

        self.sssp_installation = PseudosInstallWidget(auto_start=False)
        ipw.dlink(
            (self.sssp_installation, "busy"),
            (self._model, "installing_sssp"),
        )
        ipw.dlink(
            (self.sssp_installation, "installed"),
            (self._model, "installing_sssp"),
            lambda installed: not installed,
        )
        ipw.dlink(
            (self.sssp_installation, "installed"),
            (self._model, "sssp_installed"),
        )
        if auto_setup:
            self.sssp_installation.refresh()

    def _toggle_sssp_installation_widget(self):
        sssp_installation_display = "none" if self._model.sssp_installed else "block"
        self.sssp_installation.layout.display = sssp_installation_display

    def _update_state(self):
        if self._model.confirmed:
            self.state = self.State.SUCCESS
        elif self._model.input_structure is None:
            self.state = self.State.READY
        else:
            self.state = self.State.CONFIGURED



from __future__ import annotations

import ipywidgets as ipw
import traitlets as tl

from aiida_quantumespresso.common.types import RelaxType
from aiidalab_qe.app.parameters import DEFAULT_PARAMETERS
from aiidalab_qe.common.mixins import (
    HasInputStructure,
    HasModels,
)
from aiidalab_qe.common.panel import ConfigurationSettingsModel
from aiidalab_qe.common.wizard import QeConfirmableWizardStepModel

DEFAULT: dict = DEFAULT_PARAMETERS  # type: ignore

NO_RELAXATION_OPTION = ("Structure as is", "none")


class ConfigurationStepModel(
    QeConfirmableWizardStepModel,
    HasModels[ConfigurationSettingsModel],
    HasInputStructure,
):
    identifier = "configuration"

    relax_type_help = tl.Unicode()
    relax_type_options = tl.List([NO_RELAXATION_OPTION])
    relax_type = tl.Unicode(NO_RELAXATION_OPTION[-1], allow_none=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._default_models = {
            "workchain",
            "advanced",
        }

        self.relax_type_help_template = """
            <div style="line-height: 140%; padding-top: 0px; padding-bottom: 5px">
                You have {option_count} options:
                <br>
                (1) Structure as is: perform a self consistent calculation using
                the structure provided as input.
                <br>
                (2) Atomic positions: perform a full relaxation of the internal
                atomic coordinates.
                {full_relaxation_option}
            </div>
        """

    def update(self):
        if self.has_pbc:
            relax_type_help = self.relax_type_help_template.format(
                option_count="three",
                full_relaxation_option=(
                    """
                    <br>
                    (3) Full geometry: perform a full relaxation of the internal atomic
                    coordinates and the cell parameters.
                    """
                ),
            )
            relax_type_options = [
                NO_RELAXATION_OPTION,
                ("Atomic positions", "positions"),
                ("Full geometry", "positions_cell"),
            ]
        else:
            relax_type_help = self.relax_type_help_template.format(
                option_count="two",
                full_relaxation_option="",
            )
            relax_type_options = [
                NO_RELAXATION_OPTION,
                ("Atomic positions", "positions"),
            ]

        default = DEFAULT["workchain"]["relax_type"]
        default_available = [default in [option[1] for option in relax_type_options]]
        relax_type = default if default_available else relax_type_options[-1][-1]

        self._defaults = {
            "relax_type_help": relax_type_help,
            "relax_type_options": relax_type_options,
            "relax_type": relax_type,
        }

        self.relax_type_help = self._get_default_relax_type_help()
        self.relax_type_options = self._get_default_relax_type_options()
        self.relax_type = self._get_default_relax_type()

    def get_model_state(self):
        parameters = {
            identifier: model.get_model_state()
            for identifier, model in self.get_models()
            if model.include
        }
        parameters["workchain"] |= {
            "relax_type": self.relax_type,
            "properties": self._get_properties(),
        }
        return parameters

    def set_model_state(self, parameters):
        with self.hold_trait_notifications():
            workchain_parameters: dict = parameters.get("workchain", {})
            self.relax_type = workchain_parameters.get("relax_type")
            properties = set(workchain_parameters.get("properties", []))
            for identifier, model in self.get_models():
                model.include = identifier in self._default_models | properties
                if parameters.get(identifier):
                    model.set_model_state(parameters[identifier])
                    model.loaded_from_process = True

    def reset(self):
        self.confirmed = False
        self.relax_type_help = self._get_default_relax_type_help()
        self.relax_type_options = self._get_default_relax_type_options()
        self.relax_type = self._get_default_relax_type()
        for identifier, model in self.get_models():
            if identifier not in self._default_models:
                model.include = False

    def _link_model(self, model: ConfigurationSettingsModel):
        ipw.link(
            (self, "confirmed"),
            (model, "confirmed"),
        )
        super()._link_model(model)

    def _get_properties(self):
        properties = []
        for identifier, model in self.get_models():
            if identifier in self._default_models:
                continue
            if model.include:
                properties.append(identifier)
        if RelaxType(self.relax_type) is not RelaxType.NONE or not properties:
            properties.append("relax")
        return properties

    def _get_default_relax_type_help(self):
        return self._defaults.get("relax_type_help", "")

    def _get_default_relax_type_options(self):
        return self._defaults.get("relax_type_options", [NO_RELAXATION_OPTION])

    def _get_default_relax_type(self):
        options = self._get_default_relax_type_options()
        relax_type = self._defaults.get("relax_type", NO_RELAXATION_OPTION[-1])
        return (
            relax_type
            if relax_type in [option[1] for option in options]
            else options[-1][-1]
        )



"""Widgets for the submission of bands work chains.

Authors: AiiDAlab team
"""

import ipywidgets as ipw

from aiidalab_qe.app.configuration.basic.model import BasicConfigurationSettingsModel
from aiidalab_qe.common.infobox import InAppGuide
from aiidalab_qe.common.panel import ConfigurationSettingsPanel


class BasicConfigurationSettingsPanel(
    ConfigurationSettingsPanel[BasicConfigurationSettingsModel],
):
    def __init__(self, model: BasicConfigurationSettingsModel, **kwargs):
        super().__init__(model, **kwargs)
        self._model.observe(
            self._on_input_structure_change,
            "input_structure",
        )

    def render(self):
        if self.rendered:
            return

        # ElectronicType: electronic properties of material
        self.electronic_type = ipw.ToggleButtons(style={"description_width": "initial"})
        ipw.dlink(
            (self._model, "electronic_type_options"),
            (self.electronic_type, "options"),
        )
        ipw.link(
            (self._model, "electronic_type"),
            (self.electronic_type, "value"),
        )
        self.electronic_type.observe(
            self._on_spin_type_change,
            "value",
        )

        # SpinType: magnetic properties of material
        self.spin_type = ipw.ToggleButtons(style={"description_width": "initial"})
        ipw.dlink(
            (self._model, "spin_type_options"),
            (self.spin_type, "options"),
        )
        ipw.link(
            (self._model, "spin_type"),
            (self.spin_type, "value"),
        )
        self.spin_type.observe(
            self._on_spin_type_change,
            "value",
        )

        self.starting_magmom_warning = ipw.HTML(
            value="""
                <div class="alert alert-inline alert-info" style="margin-left: 10px;">
                    <b>Note:</b> only total magnetization can be set for insulators
                </div>
            """,
            layout=ipw.Layout(display="none"),
        )

        self.magnetization_info = ipw.HTML(
            value="""
                <div class="alert alert-inline alert-info" style="margin-left: 10px;">
                    <b>Note:</b> set the desired magnetic configuration in <b>advanced
                    </b> settings
                </div>
            """,
            layout=ipw.Layout(display="none"),
        )

        # Spin-Orbit calculation
        self.spin_orbit = ipw.ToggleButtons(style={"description_width": "initial"})
        ipw.dlink(
            (self._model, "spin_orbit_options"),
            (self.spin_orbit, "options"),
        )
        ipw.link(
            (self._model, "spin_orbit"),
            (self.spin_orbit, "value"),
        )

        # Work chain protocol
        self.protocol = ipw.ToggleButtons()
        ipw.dlink(
            (self._model, "protocol_options"),
            (self.protocol, "options"),
        )
        ipw.link(
            (self._model, "protocol"),
            (self.protocol, "value"),
        )

        self.warning = ipw.HTML(
            value="""
                <div
                    class="alert alert-warning"
                    style="line-height: 140%; margin: 10px 0 0"
                >
                    <p>
                        <b>Warning:</b> detected multiples atoms with different tags.
                        You may be interested in an antiferromagnetic system. Note that
                        default starting magnetic moments do not distinguish tagged
                        atoms and are set to the same value.
                    </p>
                    <p>
                        Please go to <b>Advanced settings</b> and override the default
                        values, specifying appropriate magnetic moments for each
                        species (e.g. with different signs for an antiferromagnetic
                        configuration).
                    </p>
                </div>
            """,
            layout=ipw.Layout(display="none"),
        )

        self.children = [
            InAppGuide(identifier="basic-settings"),
            ipw.HTML("""
                <div style="line-height: 140%; padding-top: 10px; padding-bottom: 10px">
                    Below you can indicate the following:
                    <ol>
                        <li>
                            If the material should be treated as an insulator or a metal
                            (if in doubt, choose "Metal")
                        </li>
                        <li>
                            If the material should be studied with magnetization/spin
                            polarization (at least twice as costly if activated)
                        </li>
                        <li>
                            If the material should be studied with spin-orbit coupling
                        </li>
                        <li>
                            The protocol to use for the calculation, which sets default
                            values balancing the accuracy and speed of the calculation
                        </li>
                    </ol>
                </div>
            """),
            ipw.HBox(
                children=[
                    ipw.Label(
                        "Electronic type:",
                        layout=ipw.Layout(justify_content="flex-start", width="120px"),
                    ),
                    self.electronic_type,
                    self.starting_magmom_warning,
                ],
                layout=ipw.Layout(align_items="baseline"),
            ),
            ipw.HBox(
                children=[
                    ipw.Label(
                        "Magnetism:",
                        layout=ipw.Layout(justify_content="flex-start", width="120px"),
                    ),
                    self.spin_type,
                    self.magnetization_info,
                ],
                layout=ipw.Layout(align_items="baseline"),
            ),
            ipw.HBox(
                children=[
                    ipw.Label(
                        "Spin-orbit coupling:",
                        layout=ipw.Layout(justify_content="flex-start", width="120px"),
                    ),
                    self.spin_orbit,
                ],
                layout=ipw.Layout(align_items="baseline"),
            ),
            ipw.HBox(
                children=[
                    ipw.Label(
                        "Protocol:",
                        layout=ipw.Layout(justify_content="flex-start", width="120px"),
                    ),
                    self.protocol,
                ],
                layout=ipw.Layout(align_items="baseline"),
            ),
            ipw.HTML("""
                <div style="line-height: 140%; padding-top: 6px; padding-bottom: 0px">
                    The "moderate" protocol represents a trade-off between accuracy and
                    speed. Choose the "fast" protocol for a faster calculation with
                    less precision and the "precise" protocol to aim at best accuracy
                    (at the price of longer/costlier calculations).
                </div>
            """),
            self.warning,
        ]

        self.rendered = True

    def _on_input_structure_change(self, _):
        self.refresh(specific="structure")

    def _on_electronic_type_change(self, _):
        self._update_info_warning_messages()

    def _on_spin_type_change(self, _):
        self._update_info_warning_messages()

    def _update_info_warning_messages(self):
        if self._model.spin_type == "collinear":
            self.magnetization_info.layout.display = "block"
            self.starting_magmom_warning.layout.display = (
                "block" if self._model.electronic_type == "insulator" else "none"
            )
            if self._model.has_tags:
                self.warning.layout.display = "flex"
        else:
            self.starting_magmom_warning.layout.display = "none"
            self.magnetization_info.layout.display = "none"
            self.warning.layout.display = "none"



from aiidalab_qe.common.panel import PluginOutline

from .model import BandsConfigurationSettingsModel
from .resources import BandsResourceSettingsModel, BandsResourceSettingsPanel
from .setting import BandsConfigurationSettingsPanel
from .workchain import workchain_and_builder


class BandsPluginOutline(PluginOutline):
    title = "Electronic band structure"


bands = {
    "outline": BandsPluginOutline,
    "configuration": {
        "panel": BandsConfigurationSettingsPanel,
        "model": BandsConfigurationSettingsModel,
    },
    "resources": {
        "panel": BandsResourceSettingsPanel,
        "model": BandsResourceSettingsModel,
    },
    "workchain": workchain_and_builder,
}



from aiidalab_qe.common.panel import PluginOutline

from .model import PdosConfigurationSettingsModel
from .resources import PdosResourceSettingsModel, PdosResourceSettingsPanel
from .setting import PdosConfigurationSettingPanel
from .workchain import workchain_and_builder


class PdosPluginOutline(PluginOutline):
    title = "Electronic projected density of states (PDOS)"


pdos = {
    "outline": PdosPluginOutline,
    "configuration": {
        "panel": PdosConfigurationSettingPanel,
        "model": PdosConfigurationSettingsModel,
    },
    "resources": {
        "panel": PdosResourceSettingsPanel,
        "model": PdosResourceSettingsModel,
    },
    "workchain": workchain_and_builder,
}



import traitlets as tl

from aiidalab_qe.app.parameters import DEFAULT_PARAMETERS
from aiidalab_qe.common.mixins import HasInputStructure
from aiidalab_qe.common.panel import ConfigurationSettingsModel

DEFAULT: dict = DEFAULT_PARAMETERS  # type: ignore


class BasicConfigurationSettingsModel(
    ConfigurationSettingsModel,
    HasInputStructure,
):
    title = "Basic settings"
    identifier = "workchain"

    dependencies = [
        "input_structure",
    ]

    protocol_options = tl.List(
        trait=tl.Tuple(tl.Unicode(), tl.Unicode()),
        default_value=[
            ("Fast", "fast"),
            ("Moderate", "moderate"),
            ("Precise", "precise"),
        ],
    )
    protocol = tl.Unicode(DEFAULT["workchain"]["protocol"])
    spin_type_options = tl.List(
        trait=tl.List(tl.Unicode()),
        default_value=[
            ["Off", "none"],
            ["On", "collinear"],
        ],
    )
    spin_type = tl.Unicode(DEFAULT["workchain"]["spin_type"])
    electronic_type_options = tl.List(
        trait=tl.List(tl.Unicode()),
        default_value=[
            ["Metal", "metal"],
            ["Insulator", "insulator"],
        ],
    )
    electronic_type = tl.Unicode(DEFAULT["workchain"]["electronic_type"])
    spin_orbit_options = tl.List(
        trait=tl.List(tl.Unicode()),
        default_value=[
            ["Off", "wo_soc"],
            ["On", "soc"],
        ],
    )
    spin_orbit = tl.Unicode("wo_soc")

    include = True

    def get_model_state(self):
        return {
            "protocol": self.protocol,
            "spin_type": self.spin_type,
            "electronic_type": self.electronic_type,
        }

    def set_model_state(self, parameters):
        self.protocol = parameters.get("protocol", self.protocol)
        self.spin_type = parameters.get("spin_type", self.spin_type)
        self.electronic_type = parameters.get("electronic_type", self.electronic_type)

    def reset(self):
        with self.hold_trait_notifications():
            self.protocol = self.traits()["protocol"].default_value
            self.spin_type = self.traits()["spin_type"].default_value
            self.electronic_type = self.traits()["electronic_type"].default_value
            self.spin_orbit = self.traits()["spin_orbit"].default_value



from __future__ import annotations

import ipywidgets as ipw

from aiidalab_qe.app.result.components import ResultsComponent
from aiidalab_qe.common.process.tree import (
    SimplifiedProcessTree,
    SimplifiedProcessTreeModel,
)
from aiidalab_widgets_base import ProcessNodesTreeWidget
from aiidalab_widgets_base.viewers import AiidaNodeViewWidget

from .model import WorkChainStatusModel


class WorkChainStatusPanel(ResultsComponent[WorkChainStatusModel]):
    def _render(self):
        model = SimplifiedProcessTreeModel()
        self.simplified_process_tree = SimplifiedProcessTree(model=model)
        ipw.dlink(
            (self._model, "process_uuid"),
            (model, "process_uuid"),
        )
        ipw.dlink(
            (self._model, "monitor_counter"),
            (model, "monitor_counter"),
        )
        model.observe(
            self._on_calculation_link_click,
            "clicked",
        )

        self.process_tree = ProcessNodesTreeWidget()
        ipw.dlink(
            (self._model, "process_uuid"),
            (self.process_tree, "value"),
        )

        self.reset_button = ipw.Button(
            description="Reset to root",
            button_style="warning",
            icon="refresh",
            tooltip="Reseed the process tree with the root node",
            layout=ipw.Layout(width="fit-content"),
        )
        self.reset_button.on_click(self._reset_process_tree)

        self.node_view = AiidaNodeViewWidget()
        ipw.dlink(
            (self.process_tree, "selected_nodes"),
            (self.node_view, "node"),
            transform=lambda nodes: nodes[0] if nodes else None,
        )

        self.to_advanced_view_button = ipw.Button(
            description="View in advanced panel",
            button_style="primary",
            icon="eye",
            tooltip="Switch to the advanced view",
            layout=ipw.Layout(width="fit-content"),
        )
        self.to_advanced_view_button.on_click(self._switch_to_advanced_view)

        simplified_tree_container = ipw.VBox(
            children=[
                self.simplified_process_tree,
            ],
        )

        simplified_tree_node_view_container = ipw.VBox(
            children=[
                self.to_advanced_view_button,
                self.node_view,
            ],
        )

        simplified_view = ipw.Box(
            children=[
                simplified_tree_container,
                simplified_tree_node_view_container,
            ]
        )
        simplified_view.add_class("simplified-view")

        advanced_view = ipw.VBox(
            children=[
                self.reset_button,
                self.process_tree,
                self.node_view,
            ],
        )
        advanced_view.add_class("advanced-view")

        self.accordion = ipw.Accordion(
            children=[
                simplified_view,
                advanced_view,
            ],
            selected_index=None,
        )
        titles = [
            "Status overview",
            "Advanced status view",
        ]
        for i, title in enumerate(titles):
            self.accordion.set_title(i, title)

        self.accordion.observe(
            self._on_accordion_change,
            "selected_index",
        )

        self.accordion.selected_index = 0

        self.children = [self.accordion]

    def _post_render(self):
        self._select_tree_root()

    def _on_monitor_counter_change(self, _):
        if self.rendered:
            self.process_tree.update()

    def _on_accordion_change(self, change):
        if change["new"] == 0:
            self.simplified_process_tree.render()

    def _on_calculation_link_click(self, change):
        if selected_node_uuid := change["new"]:
            self.process_tree.value = selected_node_uuid

    def _switch_to_advanced_view(self, _):
        self.accordion.selected_index = 1

    def _select_tree_root(self):
        if self.rendered:
            self.process_tree.value = None
            self.process_tree.value = self._model.process_uuid

    def _reset_process_tree(self, _):
        if not self.rendered:
            return
        self.process_tree.value = self._model.process_uuid



from __future__ import annotations

import traitlets as tl
from ase.formula import Formula

from aiida import orm
from aiidalab_qe.common.panel import ResultsModel
from aiidalab_qe.common.time import format_time, relative_time


class StructureResultsModel(ResultsModel):
    title = "Structure"
    identifier = "structure"

    structure = tl.Instance(orm.StructureData, allow_none=True)
    selected_view = tl.Unicode("initial")
    header = tl.Unicode()
    sub_header = tl.Unicode()
    source = tl.Instance(orm.utils.managers.NodeLinksManager, allow_none=True)
    info = tl.Unicode()
    table_data = tl.List(tl.List())

    _this_process_label = "PwRelaxWorkChain"

    HEADER_TEMPLATE = "<h2 style='margin: 0;'>{content}</h2>"
    _SUB_HEADER_TEMPLATE = "<h4 style='margin: 0; font-size: 18px'>{content}</h4>"

    @property
    def include(self):
        return True

    @property
    def is_relaxed(self):
        return "relax" in self.properties

    def update(self):
        super().update()
        with self.hold_trait_notifications():
            if not self.is_relaxed or self.selected_view == "initial":
                self.sub_header = self._SUB_HEADER_TEMPLATE.format(content="Initial")
                self.source = self.inputs
            else:
                self.sub_header = self._SUB_HEADER_TEMPLATE.format(content="Relaxed")
                self.source = self.outputs
            self.structure = self._get_structure()
            if self.structure:
                self.info = self._get_structure_info()
                self.table_data = self._get_atom_table_data()
                if not self.header:
                    formatted = Formula(self.structure.get_formula()).format("html")
                    self.header = self.HEADER_TEMPLATE.format(content=formatted)

    def toggle_selected_view(self):
        self.selected_view = "relaxed" if self.selected_view == "initial" else "initial"

    def _get_structure(self) -> orm.StructureData | None:
        try:
            return self.source.structure if self.source else None
        except AttributeError:
            # If source is outputs but job failed, there may not be a structure
            return None

    def _get_structure_info(self):
        structure = self.structure
        formatted = format_time(structure.ctime)
        relative = relative_time(structure.ctime)
        return f"""
            <div style='line-height: 1.4;'>
                <strong>PK:</strong> {structure.pk}<br>
                <strong>Label:</strong> {structure.label}<br>
                <strong>Description:</strong> {structure.description}<br>
                <strong>Number of atoms:</strong> {len(structure.sites)}<br>
                <strong>Creation time:</strong> {formatted} ({relative})<br>
            </div>
        """

    def _get_atom_table_data(self):
        structure = self.structure.get_ase()
        data = [
            [
                "Atom index",
                "Chemical symbol",
                "Tag",
                "x (Å)",
                "y (Å)",
                "z (Å)",
            ]
        ]
        positions = structure.positions
        chemical_symbols = structure.get_chemical_symbols()
        tags = structure.get_tags()

        for index, (symbol, tag, position) in enumerate(
            zip(chemical_symbols, tags, positions), start=1
        ):
            formatted_position = [f"{coord:.2f}" for coord in position]
            data.append([index, symbol, tag, *formatted_position])

        return data



"""Panel for Bands plugin."""

import ipywidgets as ipw

from aiidalab_qe.common.infobox import InAppGuide
from aiidalab_qe.common.panel import ConfigurationSettingsPanel
from aiidalab_qe.plugins.bands.model import BandsConfigurationSettingsModel


class BandsConfigurationSettingsPanel(
    ConfigurationSettingsPanel[BandsConfigurationSettingsModel],
):
    def render(self):
        if self.rendered:
            return

        self.projwfc_bands = ipw.Checkbox(
            indent=False,
            description="Fat bands calculation",
            style={"description_width": "initial"},
        )
        ipw.link(
            (self._model, "projwfc_bands"),
            (self.projwfc_bands, "value"),
        )

        self.children = [
            InAppGuide(identifier="bands-settings"),
            ipw.HTML("""
                <div style="line-height: 140%;">
                    <p style="margin-bottom: 10px;">
                        The band structure workflow will automatically detect the
                        default path in reciprocal space using the
                        <a
                            href="https://www.materialscloud.org/work/tools/seekpath"
                            target="_blank"
                        >SeeK-path tool</a>.
                    </p>
                    <p>
                        Fat Bands is a band structure plot that includes the angular
                        momentum contributions from specific atoms or orbitals to each
                        energy band. The thickness of the bands represents the strength
                        of these contributions, providing insight into the electronic
                        structure.
                    </p>
                </div>
            """),
            self.projwfc_bands,
        ]

        self.rendered = True



import traitlets as tl

from aiidalab_qe.common.panel import ConfigurationSettingsModel


class BandsConfigurationSettingsModel(ConfigurationSettingsModel):
    """Model for the band structure plugin."""

    title = "Band structure"
    identifier = "bands"

    projwfc_bands = tl.Bool(False)

    def get_model_state(self):
        return {"projwfc_bands": self.projwfc_bands}

    def set_model_state(self, parameters: dict):
        self.projwfc_bands = parameters.get("projwfc_bands", False)

    def reset(self):
        self.projwfc_bands = False



import traitlets as tl

from aiidalab_qe.common.mixins import HasInputStructure
from aiidalab_qe.common.wizard import QeConfirmableWizardStepModel


class StructureStepModel(
    QeConfirmableWizardStepModel,
    HasInputStructure,
):
    identifier = "structure"

    structure_name = tl.Unicode("")
    manager_output = tl.Unicode("")

    installing_sssp = tl.Bool(False)
    sssp_installed = tl.Bool(allow_none=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.confirmation_exceptions += [
            "installing_sssp",
            "sssp_installed",
        ]

    def update_widget_text(self):
        if not self.has_structure:
            self.structure_name = ""
        else:
            self.manager_output = ""
            self.structure_name = str(self.input_structure.get_formula())

    def reset(self):
        self.input_structure = None
        self.structure_name = ""
        self.manager_output = ""

    def _check_blockers(self):
        if not self.sssp_installed:
            yield "The SSSP library is not installed"



from __future__ import annotations

from pathlib import Path

import traitlets as tl
from bs4 import BeautifulSoup, PageElement

import aiidalab_qe
from aiidalab_qe.app.utils import get_entry_items


class GuideManager(tl.HasTraits):
    """A global guide manager that loads and manages guide sections."""

    active_guide = tl.Unicode("none", allow_none=True)

    def __init__(self, *args, **kwargs):
        """`GuideManager` constructor."""

        super().__init__(*args, **kwargs)
        guides = Path(aiidalab_qe.__file__).parent.joinpath("guides").glob("*")
        self._guides = {
            "general": {
                guide.stem.split("_", maxsplit=1)[1]: guide.absolute()
                for guide in sorted(guides, key=lambda x: x.stem.split("_")[0])
            }
        }

        self._fetch_plugin_guides()

        self.content = BeautifulSoup()

        self.observe(
            self._on_active_guide_change,
            "active_guide",
        )

    @property
    def has_guide(self) -> bool:
        return self.active_guide != "none"

    def get_guide_categories(self) -> list[str]:
        """Return a list of available guides.

        Returns
        -------
        `list[str]`
            A list of the names of available guides.
        """
        return [*self._guides.keys()]

    def get_guides(self, identifier: str) -> list[str]:
        """Return a list of available sub-guides.

        Returns
        -------
        `list[str]`
            A list of the names of available sub-guides.
        """
        return [*self._guides[identifier].keys()] if identifier != "none" else []

    def get_guide_section_by_id(self, content_id: str) -> PageElement | None:
        """Return a guide section by its HTML `id` attribute.

        Parameters
        ----------
        `content_id` : `str`
            The HTML `id` attribute of the guide section.

        Returns
        -------
        `PageElement` | `None`
            The guide section or `None` if not found.
        """
        return self.content.find(attrs={"id": content_id})  # type: ignore

    def _on_active_guide_change(self, _):
        """Load the contents of the active guide."""
        if self.active_guide == "none":
            self.content = BeautifulSoup()
            return
        category, guide = self.active_guide.split("/")
        guide_path = self._guides[category][guide]
        html = Path(guide_path).read_text() if guide_path else ""
        self.content = BeautifulSoup(html, "html.parser")

    def _fetch_plugin_guides(self):
        """Fetch guides from plugins."""
        entries: dict = get_entry_items("aiidalab_qe.properties", "guides")
        for identifier, guides in entries.items():
            path = Path(guides)
            if identifier not in self._guides:
                self._guides[identifier] = {}
            for guide in sorted(path.glob("*"), key=lambda x: x.stem.split("_")[0]):
                stem = guide.stem.split("_", maxsplit=1)[1]
                self._guides[identifier][stem] = guide.absolute()


guide_manager = GuideManager()



"""Class to .

Authors:

    AiiDAlab Team
"""

from __future__ import annotations

import os
import typing as t

import ipywidgets as ipw
import traitlets as tl

from aiida import orm
from aiida.common.extendeddicts import AttributeDict
from aiidalab_qe.app.parameters import DEFAULT_PARAMETERS
from aiidalab_qe.common.code.model import CodeModel
from aiidalab_qe.common.infobox import InAppGuide
from aiidalab_qe.common.mixins import Confirmable, HasBlockers, HasModels, HasProcess
from aiidalab_qe.common.mvc import Model
from aiidalab_qe.common.widgets import (
    LoadingWidget,
    PwCodeResourceSetupWidget,
    QEAppComputationalResourcesWidget,
)

DEFAULT: dict = DEFAULT_PARAMETERS  # type: ignore


class PanelModel(Model):
    """Base class for all panel models.

    Attributes
    ----------
    `title` : `str`
        The title to be shown in the GUI.
    `identifier` : `str`
        Which plugin this panel belong to.
    """

    title = ""
    identifier = ""


PM = t.TypeVar("PM", bound=PanelModel)


class Panel(ipw.VBox, t.Generic[PM]):
    """Base class for all panels."""

    rendered = False
    loading_message = "Loading {identifier} panel"

    def __init__(self, model: PM, **kwargs):
        loading_message = self.loading_message.format(identifier=model.identifier)
        loading_message = loading_message.replace("_", " ")
        self.loading_message = LoadingWidget(loading_message)
        super().__init__(children=[self.loading_message], **kwargs)
        self._model = model

    def render(self):
        raise NotImplementedError()


class PluginOutline(ipw.HBox):
    title = "Outline"
    description = ""

    def __init__(self, **kwargs):
        self.include = ipw.Checkbox(
            description=self.title,
            indent=False,
            style={"description_width": "initial"},
        )

        super().__init__(
            children=[
                self.include,
                ipw.HTML(f"""
                    <div style="line-height: 140%; padding-top: 0px; padding-bottom: 5px">
                        {self.description}
                    </div>
                """),
            ],
            **kwargs,
        )


class SettingsModel(PanelModel, HasBlockers):
    """Base model for settings models."""

    dependencies: list[str] = []

    include = tl.Bool(False)
    loaded_from_process = tl.Bool(False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._defaults = {}

    def update(self):
        """Updates the model."""
        pass

    def get_model_state(self) -> dict:
        """Retrieves the current state of the model as a dictionary."""
        raise NotImplementedError()

    def set_model_state(self, parameters: dict):
        """Distributes the parameters of a loaded calculation to the model."""
        raise NotImplementedError()

    def reset(self):
        """Resets the model to present defaults."""
        pass


SM = t.TypeVar("SM", bound=SettingsModel)


class SettingsPanel(Panel[SM]):
    """Base model for settings panels."""

    updated = False

    def __init__(self, model: SM, **kwargs):
        super().__init__(model=model, **kwargs)
        self.links = []


class ConfigurationSettingsModel(SettingsModel, Confirmable):
    """Base model for configuration settings models."""

    def update(self, specific=""):
        """Updates the model.

        Parameters
        ----------
        `specific` : `str`, optional
            If provided, specifies the level of update.
        """
        pass


CSM = t.TypeVar("CSM", bound=ConfigurationSettingsModel)


class ConfigurationSettingsPanel(SettingsPanel[CSM]):
    """Base class for configuration settings panels."""

    def refresh(self, specific=""):
        """Refreshes the settings panel.

        Unlinks the panel's widgets. If the panel's model is included in the
        calculation, also updates the model's defaults. Resets the model to
        these defaults if there is no input structure.

        Parameters
        ----------
        `specific` : `str`, optional
            If provided, specifies the level of refresh.
        """
        self.updated = False
        self._unsubscribe()
        if self._model.include:
            self.update(specific)
        if "PYTEST_CURRENT_TEST" in os.environ:
            # Skip resetting to avoid having to inject a structure when testing
            return
        if hasattr(self._model, "input_structure") and not self._model.input_structure:
            self._reset()

    def update(self, specific=""):
        """Updates the model if not yet updated.

        Parameters
        ----------
        `specific` : `str`, optional
            If provided, specifies the level of update.
        """
        if self.updated:
            return
        if not self._model.loaded_from_process:
            self._model.update(specific)
        self.updated = True

    def _unsubscribe(self):
        """Unlinks any linked widgets."""
        for link in self.links:
            link.unlink()
        self.links.clear()

    def _reset(self):
        """Resets the model to present defaults."""
        self.updated = False
        self._model.reset()


class ResourceSettingsModel(SettingsModel, HasModels[CodeModel]):
    """Base model for resource setting models."""

    global_codes = tl.Dict(
        key_trait=tl.Unicode(),
        value_trait=tl.Dict(),
    )

    warning_messages = tl.Unicode("")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Used by the code-setup thread to fetch code options
        self.DEFAULT_USER_EMAIL = orm.User.collection.get_default().email

    def add_model(self, identifier, model):
        super().add_model(identifier, model)
        model.update(self.DEFAULT_USER_EMAIL)

    def refresh_codes(self):
        for _, code_model in self.get_models():
            code_model.update(self.DEFAULT_USER_EMAIL, refresh=True)

    def get_model_state(self):
        return {
            "codes": {
                identifier: code_model.get_model_state()
                for identifier, code_model in self.get_models()
                if code_model.is_ready
            },
        }

    def set_model_state(self, parameters: dict):
        self.set_selected_codes(parameters.get("codes", {}))

    def get_selected_codes(self) -> dict[str, dict]:
        return {
            identifier: code_model.get_model_state()
            for identifier, code_model in self.get_models()
            if code_model.is_ready
        }

    def set_selected_codes(self, code_data=DEFAULT["codes"]):
        for identifier, code_model in self.get_models():
            if identifier in code_data:
                code_model.set_model_state(code_data[identifier])

    def _check_blockers(self):
        return []


RSM = t.TypeVar("RSM", bound=ResourceSettingsModel)


class ResourceSettingsPanel(SettingsPanel[RSM]):
    """Base class for resource setting panels."""

    def __init__(self, model, **kwargs):
        super().__init__(model, **kwargs)
        self.code_widgets = {}

    def _on_code_resource_change(self, _):
        pass

    def _on_code_options_change(self, change: dict):
        widget: ipw.Dropdown = change["owner"]
        widget.disabled = not widget.options

    def _toggle_code(self, code_model: CodeModel):
        if not self.rendered:
            return
        if not code_model.is_rendered:
            loading_message = LoadingWidget(f"Loading {code_model.name} code")
            self.code_widgets_container.children += (loading_message,)
        if code_model.name not in self.code_widgets:
            code_widget = code_model.code_widget_class(
                description=code_model.description,
                default_calc_job_plugin=code_model.default_calc_job_plugin,
            )
            self.code_widgets[code_model.name] = code_widget
        else:
            code_widget = self.code_widgets[code_model.name]
        if not code_model.is_rendered:
            code_widget.observe(
                code_widget.update_resources,
                "value",
            )
            self._render_code_widget(code_model, code_widget)

    def _render_code_widget(
        self,
        code_model: CodeModel,
        code_widget: QEAppComputationalResourcesWidget,
    ):
        ipw.dlink(
            (code_model, "options"),
            (code_widget.code_selection.code_select_dropdown, "options"),
        )
        ipw.link(
            (code_model, "selected"),
            (code_widget, "value"),
        )
        ipw.link(
            (code_model, "num_cpus"),
            (code_widget.num_cpus, "value"),
        )
        ipw.link(
            (code_model, "num_nodes"),
            (code_widget.num_nodes, "value"),
        )
        ipw.link(
            (code_model, "ntasks_per_node"),
            (code_widget.resource_detail.ntasks_per_node, "value"),
        )
        ipw.link(
            (code_model, "cpus_per_task"),
            (code_widget.resource_detail.cpus_per_task, "value"),
        )
        ipw.link(
            (code_model, "max_wallclock_seconds"),
            (code_widget.resource_detail.max_wallclock_seconds, "value"),
        )
        if isinstance(code_widget, PwCodeResourceSetupWidget):
            ipw.link(
                (code_model, "parallelization_override"),
                (code_widget.parallelization.override, "value"),
            )
            ipw.link(
                (code_model, "npool"),
                (code_widget.parallelization.npool, "value"),
            )
            code_model.observe(
                self._on_code_resource_change,
                [
                    "parallelization_override",
                    "npool",
                ],
            )
        code_model.observe(
            self._on_code_resource_change,
            [
                "selected",
                "num_cpus",
                "num_nodes",
                "ntasks_per_node",
                "cpus_per_task",
                "max_wallclock_seconds",
            ],
        )
        code_widget.code_selection.code_select_dropdown.observe(
            self._on_code_options_change,
            "options",
        )
        code_widgets = self.code_widgets_container.children[:-1]  # type: ignore
        self.code_widgets_container.children = [*code_widgets, code_widget]
        code_model.is_rendered = True


class PluginResourceSettingsModel(ResourceSettingsModel):
    """Base model for plugin resource setting models."""

    dependencies = [
        "global.global_codes",
    ]

    override = tl.Bool(False)

    def add_model(self, identifier, model: CodeModel):
        super().add_model(identifier, model)
        model.activate()

    def update(self):
        """Updates the code models from the global resources.

        Skips synchronization with global resources if the user has chosen to override
        the resources for the plugin codes.
        """
        if self.override:
            return
        for _, code_model in self.get_models():
            model_key = code_model.default_calc_job_plugin.replace(".", "__")
            if model_key in self.global_codes:
                code_resources: dict = self.global_codes[model_key]  # type: ignore
                code_model.set_model_state(code_resources)

    def get_model_state(self):
        return {
            "override": self.override,
            **super().get_model_state(),
        }

    def set_model_state(self, parameters: dict):
        self.override = parameters.get("override", False)
        super().set_model_state(parameters)

    def _link_model(self, model: CodeModel):
        tl.link(
            (self, "override"),
            (model, "override"),
        )


PRSM = t.TypeVar("PRSM", bound=PluginResourceSettingsModel)


class PluginResourceSettingsPanel(ResourceSettingsPanel[PRSM]):
    """Base class for plugin resource setting panels."""

    def __init__(self, model, **kwargs):
        super().__init__(model, **kwargs)

        self._model.observe(
            self._on_global_codes_change,
            "global_codes",
        )
        self._model.observe(
            self._on_override_change,
            "override",
        )

    def render(self):
        if self.rendered:
            return

        self.override_help = ipw.HTML(
            "Click to override the resource settings for this plugin."
        )
        self.override = ipw.Checkbox(
            description="",
            indent=False,
            layout=ipw.Layout(max_width="3%"),
        )
        ipw.link(
            (self._model, "override"),
            (self.override, "value"),
        )
        self.code_widgets_container = ipw.VBox()

        self.children = [
            ipw.HBox(
                children=[
                    self.override,
                    self.override_help,
                ]
            ),
            self.code_widgets_container,
        ]

        self.rendered = True

        # Render any active codes
        for _, code_model in self._model.get_models():
            if code_model.is_active:
                self._toggle_code(code_model)

        return self.code_widgets_container

    def _on_global_codes_change(self, _):
        self._model.update()

    def _on_override_change(self, _):
        self._model.update()

    def _render_code_widget(
        self,
        code_model: CodeModel,
        code_widget: QEAppComputationalResourcesWidget,
    ):
        super()._render_code_widget(code_model, code_widget)
        self._link_override_to_widget_disable(code_model, code_widget)

    def _link_override_to_widget_disable(self, code_model, code_widget):
        """Links the override attribute of the code model to the disable attribute
        of subwidgets of the code widget."""
        ipw.dlink(
            (code_model, "override"),
            (code_widget.code_selection.code_select_dropdown, "disabled"),
            lambda override: not override,
        )
        ipw.dlink(
            (code_model, "override"),
            (code_widget.num_cpus, "disabled"),
            lambda override: not override,
        )
        ipw.dlink(
            (code_model, "override"),
            (code_widget.num_nodes, "disabled"),
            lambda override: not override,
        )
        ipw.dlink(
            (code_model, "override"),
            (code_widget.btn_setup_resource_detail, "disabled"),
            lambda override: not override,
        )
        if isinstance(code_widget, PwCodeResourceSetupWidget):
            ipw.dlink(
                (code_model, "override"),
                (code_widget.parallelization.override, "disabled"),
                lambda override: not override,
            )
            ipw.dlink(
                (code_model, "override"),
                (code_widget.parallelization.npool, "disabled"),
                lambda override: not override,
            )


class ResultsModel(PanelModel, HasProcess):
    process_status_notification = tl.Unicode("")

    _this_process_label = ""
    _this_process_uuid = None

    auto_render = False
    _completed_process = False

    CSS_MAP = {
        "finished": "success",
        "failed": "danger",
        "excepted": "danger",
        "killed": "danger",
        "queued": "warning",
        "running": "info",
        "created": "info",
    }

    @property
    def include(self):
        return self.identifier in self.properties

    @property
    def has_results(self):
        node = self.fetch_child_process_node()
        return node and node.is_finished_ok

    def update(self):
        self.auto_render = self.has_results

    def update_process_status_notification(self):
        if self._completed_process:
            self.process_status_notification = ""
            return
        status = self._get_child_process_status()
        self.process_status_notification = status
        if "success" in status:
            self._completed_process = True

    def fetch_child_process_node(self, which="this") -> orm.ProcessNode | None:
        if not self.process_uuid:
            return
        which = which.lower()
        uuid = getattr(self, f"_{which}_process_uuid")
        label = getattr(self, f"_{which}_process_label")
        if not uuid:
            root = self.fetch_process_node()
            child = next((c for c in root.called if c.process_label == label), None)
            uuid = child.uuid if child else None
        return orm.load_node(uuid) if uuid else None  # type: ignore

    def _get_child_process_status(self, which="this"):
        state, exit_message = self._get_child_state_and_exit_message(which)
        if state == "waiting":
            state = "running"
        status = state.upper()
        if exit_message:
            status = f"{status} ({exit_message})"
        label = "Status" if which == "this" else f"{which.capitalize()} status"
        alert_class = f"alert-{self.CSS_MAP.get(state, 'info')}"
        return f"""
            <div class="alert {alert_class}" style="padding: 5px 10px;">
                <b>{label}:</b> {status}
            </div>
        """

    def _get_child_state_and_exit_message(self, which="this"):
        if not (
            (node := self.fetch_child_process_node(which))
            and hasattr(node, "process_state")
            and node.process_state
        ):
            return "queued", None
        if node.is_failed:
            return "failed", node.exit_message
        return node.process_state.value, None

    def _get_child_outputs(self, which="this"):
        if not (node := self.fetch_child_process_node(which)):
            outputs = super().outputs
            child = which if which != "this" else self.identifier
            return getattr(outputs, child) if child in outputs else AttributeDict({})
        return AttributeDict({key: getattr(node.outputs, key) for key in node.outputs})


RM = t.TypeVar("RM", bound=ResultsModel)


class ResultsPanel(Panel[RM]):
    """Base class for all the result panels.

    The base class has a method to load the result of the calculation.
    And a show method to display it in the panel.
    It has a update method to update the result in the panel.
    """

    loading_message = "Loading {identifier} results"

    def __init__(self, model: RM, **kwargs):
        super().__init__(model=model, **kwargs)
        self._model.observe(
            self._on_process_change,
            "process_uuid",
        )
        self._model.observe(
            self._on_monitor_counter_change,
            "monitor_counter",
        )

    def render(self):
        if self.rendered:
            if self._model.identifier == "structure":
                self._render()
            return

        if not self._model.has_process:
            return

        self.guide = InAppGuide(
            identifier=f"{self._model.identifier}-results",
            classes=["results-panel-guide"],
        )

        self.results_container = ipw.VBox()

        if self._model.auto_render:
            self.children = [
                self.guide,
                self.results_container,
            ]
            self._load_results()
        else:
            children = [self.guide]
            if (
                self._model.identifier != "structure"
                or "relax" in self._model.properties
            ):
                children.append(self._get_controls_section())
            children.append(self.results_container)
            self.children = children
            if self._model.identifier == "structure":
                self._load_results()

        self.rendered = True

    def _on_process_change(self, _):
        self._model.update()

    def _on_monitor_counter_change(self, _):
        self._model.update_process_status_notification()

    def _on_load_results_click(self, _):
        self.load_controls.children = []
        self._load_results()

    def _load_results(self):
        self.results_container.children = [self.loading_message]
        self._render()
        self._post_render()

    def _get_controls_section(self) -> ipw.VBox:
        self.process_status_notification = ipw.HTML()
        ipw.dlink(
            (self._model, "process_status_notification"),
            (self.process_status_notification, "value"),
        )

        self.load_results_button = ipw.Button(
            description="Load results",
            button_style="warning",
            tooltip="Load the results",
            icon="refresh",
        )
        ipw.dlink(
            (self._model, "monitor_counter"),
            (self.load_results_button, "disabled"),
            lambda _: not self._model.has_results,
        )
        self.load_results_button.on_click(self._on_load_results_click)

        self.load_controls = ipw.HBox(
            children=[]
            if self._model.auto_render or self._model.identifier == "structure"
            else [
                self.load_results_button,
                ipw.HTML("""
                    <div style="margin-left: 10px">
                        <b>Note:</b> Load time may vary depending on the size of the
                        calculation
                    </div>
                """),
            ]
        )

        return ipw.VBox(
            children=[
                self.process_status_notification,
                self.load_controls,
            ]
        )

    def _render(self):
        raise NotImplementedError()

    def _post_render(self):
        pass



import ipywidgets as ipw

from aiidalab_qe.common.widgets import HBoxWithUnits

from ..subsettings import AdvancedConfigurationSubSettingsPanel
from .model import HubbardConfigurationSettingsModel


class HubbardConfigurationSettingsPanel(
    AdvancedConfigurationSubSettingsPanel[HubbardConfigurationSettingsModel],
):
    def __init__(self, model: HubbardConfigurationSettingsModel, **kwargs):
        super().__init__(model, **kwargs)

        self._model.observe(
            self._on_input_structure_change,
            "input_structure",
        )
        self._model.observe(
            self._on_hubbard_activation,
            "is_active",
        )
        self._model.observe(
            self._on_eigenvalues_definition,
            "has_eigenvalues",
        )

    def render(self):
        if self.rendered:
            return

        self.activate_hubbard_checkbox = ipw.Checkbox(
            description="Define U values",
            indent=False,
        )
        ipw.link(
            (self._model, "is_active"),
            (self.activate_hubbard_checkbox, "value"),
        )

        self.eigenvalues_help = ipw.HTML("""
            <div style="line-height: 1.4; margin: 10px 0 5px;">
                For transition metals and lanthanoids, the starting eigenvalues can be defined (magnetic calculation).
                <br>
                It is useful to suggest the desired orbital occupations  when the default choice takes another path.
                <br>
                To do so, tick the checkbox below and set the desired eigenvalues to a value other than -1 (unset).
            </div>
        """)
        self.define_eigenvalues_checkbox = ipw.Checkbox(
            description="Define eigenvalues",
            indent=False,
        )
        ipw.link(
            (self._model, "has_eigenvalues"),
            (self.define_eigenvalues_checkbox, "value"),
        )

        self.eigenvalues_container = ipw.VBox(
            children=[
                self.eigenvalues_help,
                self.define_eigenvalues_checkbox,
            ]
        )

        self.hubbard_widget = ipw.VBox()
        self.eigenvalues_widget = ipw.VBox()

        self.container = ipw.VBox()

        self.children = [
            ipw.HTML("<h2>Hubbard (DFT+U)</h2>"),
            self.activate_hubbard_checkbox,
            self.container,
        ]

        self.rendered = True

        self.refresh(specific="widgets")

    def _on_input_structure_change(self, _):
        self.refresh(specific="structure")

    def _on_hubbard_activation(self, _):
        self._toggle_hubbard_widget()

    def _on_eigenvalues_definition(self, _):
        self._toggle_eigenvalues_widget()

    def _update(self, specific=""):
        if self.updated:
            return
        self._show_loading()
        if not self._model.loaded_from_process or (specific and specific != "widgets"):
            self._model.update(specific)
        self._build_hubbard_widget()
        self._toggle_hubbard_widget()
        self._toggle_eigenvalues_widget()
        self.updated = True

    def _show_loading(self):
        if self.rendered:
            self.hubbard_widget.children = [self.loading_message]

    def _build_hubbard_widget(self):
        if not self.rendered:
            return

        children = []

        for label in self._model.orbital_labels:
            float_widget = ipw.BoundedFloatText(
                description=label,
                min=0,
                max=20,
                step=0.1,
                style={"description_width": "150px"},
            )
            link = ipw.link(
                (self._model, "parameters"),
                (float_widget, "value"),
                [
                    lambda parameters, label=label: parameters.get(label, 0.0),
                    lambda value, label=label: {
                        **self._model.parameters,
                        label: value,
                    },
                ],
            )
            self.links.append(link)
            children.append(HBoxWithUnits(float_widget, "eV"))

        if self._model.needs_eigenvalues_widget:
            children.append(self.eigenvalues_container)
            self._build_eigenvalues_widget()
        else:
            self.eigenvalues_widget.children = []

        self.hubbard_widget.children = children

    def _build_eigenvalues_widget(self):
        def update(index, spin, state, symbol, value):
            eigenvalues = [*self._model.eigenvalues]
            eigenvalues[index][spin - 1][state] = [state + 1, spin, symbol, value]
            return eigenvalues

        children = []

        for (
            kind_index,
            (kind_name, num_states),
        ) in enumerate(self._model.applicable_kind_names):
            label_layout = ipw.Layout(justify_content="flex-start", width="50px")
            spin_row_layout = ipw.Layout(grid_gap="5px")
            spin_up_row = ipw.HBox(
                children=[ipw.Label("Up:", layout=label_layout)],
                layout=spin_row_layout,
            )
            spin_down_row = ipw.HBox(
                children=[ipw.Label("Down:", layout=label_layout)],
                layout=spin_row_layout,
            )

            for state_index in range(num_states):
                eigenvalues_up = ipw.Dropdown(
                    description=f"{state_index + 1}",
                    layout=ipw.Layout(width="65px"),
                    style={"description_width": "initial"},
                )
                options_link = ipw.dlink(
                    (self._model, "eigenvalue_options"),
                    (eigenvalues_up, "options"),
                )
                value_link = ipw.link(
                    (self._model, "eigenvalues"),
                    (eigenvalues_up, "value"),
                    [
                        lambda eigenvalues,
                        kind_index=kind_index,
                        state_index=state_index: str(
                            int(eigenvalues[kind_index][0][state_index][-1])
                        ),
                        lambda value,
                        kind_index=kind_index,
                        state_index=state_index,
                        kind_name=kind_name: update(
                            kind_index,
                            1,
                            state_index,
                            kind_name,
                            float(value),
                        ),
                    ],
                )
                self.links.extend([options_link, value_link])
                spin_up_row.children += (eigenvalues_up,)

                eigenvalues_down = ipw.Dropdown(
                    description=f"{state_index + 1}",
                    layout=ipw.Layout(width="65px"),
                    style={"description_width": "initial"},
                )
                options_link = ipw.dlink(
                    (self._model, "eigenvalue_options"),
                    (eigenvalues_down, "options"),
                )
                value_link = ipw.link(
                    (self._model, "eigenvalues"),
                    (eigenvalues_down, "value"),
                    [
                        lambda eigenvalues,
                        kind_index=kind_index,
                        state_index=state_index: str(
                            int(eigenvalues[kind_index][1][state_index][-1])
                        ),
                        lambda value,
                        kind_index=kind_index,
                        state_index=state_index,
                        kind_name=kind_name: update(
                            kind_index,
                            2,
                            state_index,
                            kind_name,
                            float(value),
                        ),
                    ],
                )
                self.links.extend([options_link, value_link])
                spin_down_row.children += (eigenvalues_down,)

            children.append(
                ipw.HBox(
                    [
                        ipw.Label(
                            kind_name,
                            layout=ipw.Layout(
                                justify_content="flex-start",
                                width="80px",
                            ),
                        ),
                        ipw.VBox(
                            children=[
                                spin_up_row,
                                spin_down_row,
                            ]
                        ),
                    ]
                )
            )

        self.eigenvalues_widget.children = children

    def _toggle_hubbard_widget(self):
        if not self.rendered:
            return
        self.container.children = [self.hubbard_widget] if self._model.is_active else []

    def _toggle_eigenvalues_widget(self):
        if not self.rendered:
            return
        self.eigenvalues_container.children = (
            [
                self.eigenvalues_help,
                self.define_eigenvalues_checkbox,
                self.eigenvalues_widget,
            ]
            if self._model.has_eigenvalues
            else [
                self.eigenvalues_help,
                self.define_eigenvalues_checkbox,
            ]
        )



from __future__ import annotations

from copy import deepcopy

import ipywidgets as ipw
import traitlets as tl
from IPython.display import Javascript, display

from aiida import orm
from aiida.engine import ProcessBuilderNamespace, submit
from aiida.orm.utils.serialize import serialize
from aiidalab_qe.app.parameters import DEFAULT_PARAMETERS
from aiidalab_qe.common.mixins import HasInputStructure, HasModels
from aiidalab_qe.common.panel import PluginResourceSettingsModel, ResourceSettingsModel
from aiidalab_qe.common.wizard import QeConfirmableWizardStepModel
from aiidalab_qe.workflows import QeAppWorkChain

DEFAULT: dict = DEFAULT_PARAMETERS  # type: ignore


class SubmissionStepModel(
    QeConfirmableWizardStepModel,
    HasModels[ResourceSettingsModel],
    HasInputStructure,
):
    identifier = "submission"

    input_parameters = tl.Dict()

    process_node = tl.Instance(orm.WorkChainNode, allow_none=True)
    process_label = tl.Unicode("")
    process_description = tl.Unicode("")

    warning_messages = tl.Unicode("")

    installing_qe = tl.Bool(False)
    qe_installed = tl.Bool(allow_none=True)

    plugin_overrides = tl.List(tl.Unicode())

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.confirmation_exceptions += [
            "warning_messages",
            "installing_qe",
            "qe_installed",
        ]

        self._default_models = {
            "global",
        }

        self._ALERT_MESSAGE = """
            <div class="alert alert-{alert_class} alert-dismissible">
                <a href="#" class="close" data-dismiss="alert" aria-label="close">
                    &times;
                </a>
                <strong>{message}</strong>
            </div>
        """

    def confirm(self):
        super().confirm()
        if not self.process_node:
            self._submit()

    def update(self):
        for _, model in self.get_models():
            model.update()

    def update_process_label(self):
        if not self.input_structure:
            self.process_label = ""
            return
        structure_label = (
            self.input_structure.label
            if len(self.input_structure.label) > 0
            else self.input_structure.get_formula()
        )
        workchain_data = self.input_parameters.get(
            "workchain",
            {"properties": []},
        )

        soc_parameters = (
            self.input_parameters.get("advanced", {})
            .get("pw", {})
            .get("parameters", {})
            .get("SYSTEM", {})
            .get("lspinorb", False)
        )

        soc_info = "spin-orbit coupling" if soc_parameters else ""

        properties = [p for p in workchain_data["properties"] if p != "relax"]
        relax_type = workchain_data.get("relax_type", "none")
        relax_info = "unrelaxed"
        if relax_type != "none":
            relax_info = (
                "relax: atoms+cell" if "cell" in relax_type else "relax: atoms only"
            )
        protocol_and_magnetic_info = f"{workchain_data['protocol']} protocol"
        if workchain_data["spin_type"] != "none":
            protocol_and_magnetic_info += ", magnetic"
        properties_info = f"→ {', '.join(properties)}" if properties else ""

        label_details = [
            relax_info,
            protocol_and_magnetic_info,
            soc_info,
        ]
        filtered_label_details = [detail for detail in label_details if detail]
        label = f"{structure_label} [{', '.join(filtered_label_details)}] {properties_info}".strip()

        self.process_label = label

    def update_plugin_inclusion(self):
        properties = self._get_properties()
        for identifier, model in self.get_models():
            if identifier in self._default_models:
                continue
            model.include = identifier in properties

    def update_plugin_overrides(self):
        self.plugin_overrides = [
            identifier
            for identifier, model in self.get_models()
            if isinstance(model, PluginResourceSettingsModel)
            and model.include
            and model.override
        ]

    def update_warnings(self):
        warning_messages = self._check_warnings()
        for _, model in self.get_models():
            warning_messages += model.warning_messages
        self.warning_messages = warning_messages

    def get_model_state(self) -> dict[str, dict[str, dict]]:
        parameters: dict = deepcopy(self.input_parameters)  # type: ignore
        parameters["codes"] = {
            identifier: model.get_model_state()
            for identifier, model in self.get_models()
            if model.include
        }
        return parameters

    def set_model_state(self, parameters):
        codes: dict = parameters.get("codes", {})

        if "resources" in parameters:
            resources = parameters["resources"]
            codes |= {key: {"code": value} for key, value in codes.items()}
            codes["pw"]["nodes"] = resources["num_machines"]
            codes["pw"]["cpus"] = resources["num_mpiprocs_per_machine"]
            codes["pw"]["parallelization"] = {"npool": resources["npools"]}

        workchain_parameters: dict = parameters.get("workchain", {})
        properties = set(workchain_parameters.get("properties", []))
        included = self._default_models | properties
        for identifier, model in self.get_models():
            model.include = identifier in included
            if codes.get(identifier):
                model.set_model_state(codes[identifier])
                model.loaded_from_process = True

        if self.process_node:
            self.process_label = self.process_node.label
            self.process_description = self.process_node.description
            self.loaded_from_process = True

    def get_selected_codes(self) -> dict[str, dict]:
        return {
            identifier: code_model.get_model_state()
            for identifier, code_model in self.get_model("global").get_models()
            if code_model.is_ready
        }

    def reset(self):
        with self.hold_trait_notifications():
            self.input_structure = None
            self.input_parameters = {}
            self.process_node = None
            for identifier, model in self.get_models():
                if identifier not in self._default_models:
                    model.include = False

    def _submit(self):
        parameters = self.get_model_state()
        builder = self._create_builder(parameters)

        with self.hold_trait_notifications():
            process_node = submit(builder)

            process_node.label = self.process_label
            process_node.description = self.process_description
            # since AiiDA data node may exist in the ui_parameters,
            # we serialize it to yaml
            process_node.base.extras.set("ui_parameters", serialize(parameters))
            # store the workchain name in extras, this will help to filter the workchain in the future
            process_node.base.extras.set("workchain", parameters["workchain"])  # type: ignore
            process_node.base.extras.set(
                "structure",
                self.input_structure.get_formula(),
            )
            self.process_node = process_node

            self._update_url()

    def _update_url(self):
        pk = self.process_node.pk
        display(Javascript(f"window.history.pushState(null, '', '?pk={pk}');"))

    def _link_model(self, model: ResourceSettingsModel):
        for dependency in model.dependencies:
            dependency_parts = dependency.split(".")
            if len(dependency_parts) == 1:  # from parent, e.g. input_structure
                target_model = self
                trait = dependency
            else:  # from sibling, e.g. workchain.protocol
                sibling, trait = dependency_parts
                target_model = self.get_model(sibling)
            ipw.dlink(
                (target_model, trait),
                (model, trait),
            )

    def _get_properties(self) -> list[str]:
        return self.input_parameters.get("workchain", {}).get("properties", [])

    def _create_builder(self, parameters) -> ProcessBuilderNamespace:
        builder = QeAppWorkChain.get_builder_from_protocol(
            structure=self.input_structure,
            parameters=deepcopy(parameters),  # TODO why deepcopy again?
        )

        codes = parameters["codes"]["global"]["codes"]
        if "relax" in builder:
            builder.relax.base.pw.metadata.options.resources = {
                "num_machines": codes.get("quantumespresso__pw")["nodes"],
                "num_mpiprocs_per_machine": codes.get("quantumespresso__pw")[
                    "ntasks_per_node"
                ],
                "num_cores_per_mpiproc": codes.get("quantumespresso__pw")[
                    "cpus_per_task"
                ],
            }
            mws = codes.get("quantumespresso__pw")["max_wallclock_seconds"]
            builder.relax.base.pw.metadata.options["max_wallclock_seconds"] = mws
            parallelization = codes["quantumespresso__pw"]["parallelization"]
            builder.relax.base.pw.parallelization = orm.Dict(dict=parallelization)

        return builder

    def _check_blockers(self):
        if self.installing_qe:
            yield "Installing Quantum ESPRESSO codes..."

        if not self.qe_installed:
            yield "Quantum ESPRESSO is not yet installed"

    def _check_warnings(self):
        """Check for any warnings that should be displayed to the user."""
        return ""



def test_entry_point():
    from aiidalab_qe.app.utils import get_entries

    entries = get_entries()
    assert "bands" in entries
    assert "pdos" in entries
    assert "workchain" in entries["bands"]



from __future__ import annotations

import typing as t

import ipywidgets as ipw
import numpy as np
import traitlets as tl

from aiida import orm
from aiida_quantumespresso.calculations.functions.create_kpoints_from_distance import (
    create_kpoints_from_distance,
)
from aiida_quantumespresso.workflows.pw.base import PwBaseWorkChain
from aiidalab_qe.app.parameters import DEFAULT_PARAMETERS
from aiidalab_qe.common.mixins import HasInputStructure, HasModels
from aiidalab_qe.common.panel import ConfigurationSettingsModel
from aiidalab_qe.setup.pseudos import PseudoFamily

from .subsettings import AdvancedCalculationSubSettingsModel

if t.TYPE_CHECKING:
    from .hubbard.hubbard import HubbardConfigurationSettingsModel
    from .magnetization import MagnetizationConfigurationSettingsModel
    from .pseudos.pseudos import PseudosConfigurationSettingsModel
    from .smearing import SmearingConfigurationSettingsModel

DEFAULT: dict = DEFAULT_PARAMETERS  # type: ignore


class AdvancedConfigurationSettingsModel(
    ConfigurationSettingsModel,
    HasModels[AdvancedCalculationSubSettingsModel],
    HasInputStructure,
):
    title = "Advanced settings"
    identifier = "advanced"

    dependencies = [
        "input_structure",
        "workchain.protocol",
        "workchain.spin_type",
        "workchain.electronic_type",
        "workchain.spin_orbit",
    ]

    protocol = tl.Unicode()
    spin_type = tl.Unicode()
    electronic_type = tl.Unicode()
    spin_orbit = tl.Unicode()

    clean_workdir = tl.Bool(DEFAULT["advanced"]["clean_workdir"])
    total_charge = tl.Float(DEFAULT["advanced"]["tot_charge"])
    van_der_waals_options = tl.List(
        trait=tl.List(tl.Unicode()),
        default_value=[
            ["None", "none"],
            ["Grimme-D3", "dft-d3"],
            ["Grimme-D3BJ", "dft-d3bj"],
            ["Grimme-D3M", "dft-d3m"],
            ["Grimme-D3MBJ", "dft-d3mbj"],
            ["Tkatchenko-Scheffler", "ts-vdw"],
        ],
    )
    van_der_waals = tl.Unicode(DEFAULT["advanced"]["vdw_corr"])
    forc_conv_thr = tl.Float(0.0)
    forc_conv_thr_step = tl.Float(1e-4)
    etot_conv_thr = tl.Float(0.0)
    etot_conv_thr_step = tl.Float(1e-5)
    scf_conv_thr = tl.Float(0.0)
    scf_conv_thr_step = tl.Float(1e-10)
    electron_maxstep = tl.Int(80)
    optimization_maxsteps = tl.Int(50)

    kpoints_distance = tl.Float(0.0)
    mesh_grid = tl.Unicode("")

    include = True

    dftd3_version = {
        "dft-d3": 3,
        "dft-d3bj": 4,
        "dft-d3m": 5,
        "dft-d3mbj": 6,
    }

    def update(self, specific=""):
        with self.hold_trait_notifications():
            if not specific or specific != "mesh":
                parameters = PwBaseWorkChain.get_protocol_inputs(self.protocol)
                self._update_kpoints_distance(parameters)
                if specific == "protocol":
                    self._update_thresholds(parameters)
            self._update_kpoints_mesh()

    def get_model_state(self):
        num_atoms = len(self.input_structure.sites) if self.input_structure else 1
        parameters = {
            "initial_magnetic_moments": None,
            "pw": {
                "parameters": {
                    "SYSTEM": {
                        "tot_charge": self.total_charge,
                    },
                    "CONTROL": {
                        "forc_conv_thr": self.forc_conv_thr,
                        "etot_conv_thr": self.etot_conv_thr * num_atoms,
                    },
                    "ELECTRONS": {
                        "conv_thr": self.scf_conv_thr * num_atoms,
                        "electron_maxstep": self.electron_maxstep,
                    },
                }
            },
            "clean_workdir": self.clean_workdir,
            "kpoints_distance": self.kpoints_distance,
            "optimization_maxsteps": self.optimization_maxsteps,
        }

        hubbard: HubbardConfigurationSettingsModel = self.get_model("hubbard")  # type: ignore
        if hubbard.is_active:
            parameters["hubbard_parameters"] = {
                "hubbard_u": {
                    label: value for label, value in hubbard.parameters.items() if value
                }
            }
            if hubbard.has_eigenvalues:
                parameters["pw"]["parameters"]["SYSTEM"] |= {
                    "starting_ns_eigenvalue": hubbard.get_active_eigenvalues()
                }

        pseudos: PseudosConfigurationSettingsModel = self.get_model("pseudos")  # type: ignore
        parameters["pseudo_family"] = pseudos.family
        if pseudos.dictionary:
            parameters["pw"]["pseudos"] = pseudos.dictionary
            parameters["pw"]["parameters"]["SYSTEM"]["ecutwfc"] = pseudos.ecutwfc
            parameters["pw"]["parameters"]["SYSTEM"]["ecutrho"] = pseudos.ecutrho

        if self.van_der_waals in ["none", "ts-vdw"]:
            parameters["pw"]["parameters"]["SYSTEM"]["vdw_corr"] = self.van_der_waals
        else:
            parameters["pw"]["parameters"]["SYSTEM"]["vdw_corr"] = "dft-d3"
            parameters["pw"]["parameters"]["SYSTEM"]["dftd3_version"] = (
                self.dftd3_version[self.van_der_waals]
            )

        smearing: SmearingConfigurationSettingsModel = self.get_model("smearing")  # type: ignore
        if self.electronic_type == "metal":
            # smearing type setting
            parameters["pw"]["parameters"]["SYSTEM"]["smearing"] = smearing.type
            # smearing degauss setting
            parameters["pw"]["parameters"]["SYSTEM"]["degauss"] = smearing.degauss

        magnetization: MagnetizationConfigurationSettingsModel = self.get_model(
            "magnetization"
        )  # type: ignore
        if self.spin_type == "collinear":
            parameters["initial_magnetic_moments"] = magnetization.moments

        # Set tot_magnetization for collinear simulations.
        if self.spin_type == "collinear":
            # Conditions for metallic systems.
            # Select the magnetization type and set the value
            if self.electronic_type == "metal":
                if magnetization.type == "tot_magnetization":
                    parameters["pw"]["parameters"]["SYSTEM"]["tot_magnetization"] = (
                        magnetization.total
                    )
                else:
                    parameters["initial_magnetic_moments"] = magnetization.moments
            # Conditions for insulator systems. Default value is 0.0
            elif self.electronic_type == "insulator":
                parameters["pw"]["parameters"]["SYSTEM"]["tot_magnetization"] = (
                    magnetization.total
                )

        # Spin-Orbit calculation
        if self.spin_orbit == "soc":
            parameters["pw"]["parameters"]["SYSTEM"]["lspinorb"] = True
            parameters["pw"]["parameters"]["SYSTEM"]["noncolin"] = True
            parameters["pw"]["parameters"]["SYSTEM"]["nspin"] = 4

        return parameters

    def set_model_state(self, parameters):
        pseudos: PseudosConfigurationSettingsModel = self.get_model("pseudos")  # type: ignore
        if "pseudo_family" in parameters:
            pseudo_family = PseudoFamily.from_string(parameters["pseudo_family"])
            library = pseudo_family.library
            accuracy = pseudo_family.accuracy
            pseudos.library = f"{library} {accuracy}"
            pseudos.functional = pseudo_family.functional
            pseudos.family = parameters["pseudo_family"]

        if "pseudos" in parameters["pw"]:
            pseudos.dictionary = parameters["pw"]["pseudos"]
            pseudos.ecutwfc = parameters["pw"]["parameters"]["SYSTEM"]["ecutwfc"]
            pseudos.ecutrho = parameters["pw"]["parameters"]["SYSTEM"]["ecutrho"]

        self.kpoints_distance = parameters.get("kpoints_distance", 0.15)
        self.optimization_maxsteps = parameters.get("optimization_maxsteps", 50)

        if (pw_parameters := parameters.get("pw", {}).get("parameters")) is not None:
            self._set_pw_parameters(pw_parameters)

        magnetization: MagnetizationConfigurationSettingsModel = self.get_model(
            "magnetization"
        )  # type: ignore
        if magnetic_moments := parameters.get("initial_magnetic_moments"):
            if isinstance(magnetic_moments, (int, float)):
                magnetic_moments = [magnetic_moments]
            if isinstance(magnetic_moments, list):
                magnetic_moments = dict(
                    zip(
                        self.input_structure.get_kind_names(),
                        magnetic_moments,
                    )
                )
            magnetization.moments = magnetic_moments

        hubbard: HubbardConfigurationSettingsModel = self.get_model("hubbard")  # type: ignore
        if parameters.get("hubbard_parameters"):
            hubbard.is_active = True
            hubbard.parameters = parameters["hubbard_parameters"]["hubbard_u"]
            starting_ns_eigenvalue = (
                parameters.get("pw", {})
                .get("parameters", {})
                .get("SYSTEM", {})
                .get("starting_ns_eigenvalue")
            )
            if starting_ns_eigenvalue is not None:
                hubbard.set_active_eigenvalues(starting_ns_eigenvalue)

    def reset(self):
        with self.hold_trait_notifications():
            self.total_charge = self._get_default("total_charge")
            self.van_der_waals = self._get_default("van_der_waals")
            self.forc_conv_thr = self._get_default("forc_conv_thr")
            self.forc_conv_thr_step = self._get_default("forc_conv_thr_step")
            self.etot_conv_thr = self._get_default("etot_conv_thr")
            self.etot_conv_thr_step = self._get_default("etot_conv_thr_step")
            self.scf_conv_thr = self._get_default("scf_conv_thr")
            self.scf_conv_thr_step = self._get_default("scf_conv_thr_step")
            self.electron_maxstep = self._get_default("electron_maxstep")
            self.kpoints_distance = self._get_default("kpoints_distance")
            self.optimization_maxsteps = self._get_default("optimization_maxsteps")

    def _get_default(self, trait):
        return self._defaults.get(trait, self.traits()[trait].default_value)

    def _link_model(self, model: AdvancedCalculationSubSettingsModel):
        ipw.dlink(
            (self, "loaded_from_process"),
            (model, "loaded_from_process"),
        )
        model.observe(
            self._on_any_change,
            tl.All,
        )
        super()._link_model(model)

    def _update_kpoints_mesh(self, _=None):
        if not self.has_structure:
            mesh_grid = ""
        elif self.kpoints_distance > 0:
            mesh = create_kpoints_from_distance.process_class._func(
                self.input_structure,
                orm.Float(self.kpoints_distance),
                orm.Bool(False),
            )
            mesh_grid = f"Mesh {mesh.get_kpoints_mesh()[0]!s}"
        else:
            mesh_grid = "Please select a number higher than 0.0"
        self._defaults["mesh_grid"] = mesh_grid
        self.mesh_grid = mesh_grid

    def _update_kpoints_distance(self, parameters):
        kpoints_distance = parameters["kpoints_distance"] if self.has_pbc else 100.0
        self._defaults["kpoints_distance"] = kpoints_distance
        self.kpoints_distance = self._defaults["kpoints_distance"]

    def _update_thresholds(self, parameters):
        etot_value = parameters["meta_parameters"]["etot_conv_thr_per_atom"]
        self._set_value_and_step("etot_conv_thr", etot_value)
        self.etot_conv_thr = self._defaults["etot_conv_thr"]
        self.etot_conv_thr_step = self._defaults["etot_conv_thr_step"]

        scf_value = parameters["meta_parameters"]["conv_thr_per_atom"]
        self._set_value_and_step("scf_conv_thr", scf_value)
        self.scf_conv_thr = self._defaults["scf_conv_thr"]
        self.scf_conv_thr_step = self._defaults["scf_conv_thr_step"]

        forc_value = parameters["pw"]["parameters"]["CONTROL"]["forc_conv_thr"]
        self._set_value_and_step("forc_conv_thr", forc_value)
        self.forc_conv_thr = self._defaults["forc_conv_thr"]
        self.forc_conv_thr_step = self._defaults["forc_conv_thr_step"]

    def _set_value_and_step(self, attribute, value):
        self._defaults[attribute] = value
        if value != 0:
            order_of_magnitude = np.floor(np.log10(abs(value)))
            step = 10 ** (order_of_magnitude - 1)
        else:
            step = 0.1
        self._defaults[f"{attribute}_step"] = step

    def _set_pw_parameters(self, pw_parameters):
        system_params = pw_parameters.get("SYSTEM", {})
        control_params = pw_parameters.get("CONTROL", {})
        electron_params = pw_parameters.get("ELECTRONS", {})

        num_atoms = len(self.input_structure.sites) if self.input_structure else 1

        self.forc_conv_thr = control_params.get("forc_conv_thr", 0.0)
        self.etot_conv_thr = control_params.get("etot_conv_thr", 0.0) / num_atoms
        self.scf_conv_thr = electron_params.get("conv_thr", 0.0) / num_atoms
        self.electron_maxstep = electron_params.get("electron_maxstep", 80)

        self.total_charge = system_params.get("tot_charge", 0)
        self.spin_orbit = "soc" if "lspinorb" in system_params else "wo_soc"

        self.van_der_waals = self.dftd3_version.get(
            system_params.get("dftd3_version"),
            system_params.get("vdw_corr", "none"),
        )

        smearing: SmearingConfigurationSettingsModel = self.get_model("smearing")  # type: ignore
        if "degauss" in system_params:
            smearing.degauss = system_params["degauss"]

        if "smearing" in system_params:
            smearing.type = system_params["smearing"]

        magnetization: MagnetizationConfigurationSettingsModel = self.get_model(
            "magnetization"
        )  # type: ignore
        if "tot_magnetization" in system_params:
            magnetization.type = "tot_magnetization"
            magnetization.total = system_params["tot_magnetization"]



from datetime import datetime

from dateutil.relativedelta import relativedelta


def format_time(time: datetime):
    return time.strftime("%Y-%m-%d %H:%M:%S")


def relative_time(time: datetime):
    # TODO consider using humanize or arrow libraries for this
    now = datetime.now(time.tzinfo)
    delta = relativedelta(now, time)
    if delta.years > 0:
        return f"{delta.years} year{'s' if delta.years > 1 else ''} ago"
    elif delta.months > 0:
        return f"{delta.months} month{'s' if delta.months > 1 else ''} ago"
    elif delta.days > 0:
        return f"{delta.days} day{'s' if delta.days > 1 else ''} ago"
    elif delta.hours > 0:
        return f"{delta.hours} hour{'s' if delta.hours > 1 else ''} ago"
    elif delta.minutes > 0:
        return f"{delta.minutes} minute{'s' if delta.minutes > 1 else ''} ago"
    else:
        return f"{delta.seconds} second{'s' if delta.seconds > 1 else ''} ago"



import os
import typing as t

import ipywidgets as ipw
import traitlets as tl

from aiidalab_qe.common.mvc import Model


class AdvancedCalculationSubSettingsModel(Model):
    identifier = "sub"
    dependencies = []

    loaded_from_process = tl.Bool(False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._defaults = {}

    def update(self, specific=""):
        """Updates the model.

        Parameters
        ----------
        `specific` : `str`, optional
            If provided, specifies the level of update.
        """
        raise NotImplementedError()

    def reset(self):
        """Resets the model to present defaults."""
        raise NotImplementedError()


M = t.TypeVar("M", bound=AdvancedCalculationSubSettingsModel)


class AdvancedConfigurationSubSettingsPanel(ipw.VBox, t.Generic[M]):
    def __init__(self, model: M, **kwargs):
        from aiidalab_qe.common.widgets import LoadingWidget

        self.loading_message = LoadingWidget(f"Loading {model.identifier} settings")

        super().__init__(
            layout={"justify_content": "space-between", **kwargs.get("layout", {})},
            children=[self.loading_message],
            **kwargs,
        )

        self._model = model

        self.rendered = False
        self.updated = False

        self.links = []

    def render(self):
        raise NotImplementedError()

    def refresh(self, specific=""):
        """Refreshes the subsettings section.

        Unlinks any linked widgets and updates the model's defaults.
        Resets the model to these defaults if there is no input structure.

        Parameters
        ----------
        `specific` : `str`, optional
            If provided, specifies the level of refresh.
        """
        self.updated = False
        self._unsubscribe()
        self._update(specific)
        if "PYTEST_CURRENT_TEST" in os.environ:
            # Skip resetting to avoid having to inject a structure when testing
            return
        if hasattr(self._model, "input_structure") and not self._model.input_structure:
            self._reset()

    def _update(self, specific=""):
        """Updates the model if not yet updated.

        Parameters
        ----------
        `specific` : `str`, optional
            If provided, specifies the level of update.
        """
        if self.updated:
            return
        if not self._model.loaded_from_process:
            self._model.update(specific)
        self.updated = True

    def _unsubscribe(self):
        """Unlinks any linked widgets."""
        for link in self.links:
            link.unlink()
        self.links.clear()

    def _reset(self):
        """Resets the model to present defaults."""
        self.updated = False
        self._model.reset()



# trigger registration of the viewer widget:
from .node_view import CalcJobNodeViewerWidget  # noqa: F401
from .process import QeAppWorkChainSelector, WorkChainSelector
from .widgets import (
    AddingTagsEditor,
    LazyLoadedOptimade,
    LazyLoadedStructureBrowser,
    PeriodicityEditor,
    ShakeNBreakEditor,
)

__all__ = [
    "AddingTagsEditor",
    "LazyLoadedOptimade",
    "LazyLoadedStructureBrowser",
    "PeriodicityEditor",
    "QeAppWorkChainSelector",
    "WorkChainSelector",
    "ShakeNBreakEditor",
]



from .advanced import AdvancedConfigurationSettingsPanel
from .model import AdvancedConfigurationSettingsModel

__all__ = [
    "AdvancedConfigurationSettingsModel",
    "AdvancedConfigurationSettingsPanel",
]



"""Panel for Bands plugin."""

from aiidalab_qe.common.code.model import CodeModel, PwCodeModel
from aiidalab_qe.common.panel import (
    PluginResourceSettingsModel,
    PluginResourceSettingsPanel,
)


class BandsResourceSettingsModel(PluginResourceSettingsModel):
    """Model for the band structure plugin."""

    title = "Band structure"
    identifier = "bands"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.add_models(
            {
                "pw": PwCodeModel(
                    name="pw.x",
                    description="pw.x",
                    default_calc_job_plugin="quantumespresso.pw",
                ),
                "projwfc_bands": CodeModel(
                    name="projwfc.x",
                    description="projwfc.x",
                    default_calc_job_plugin="quantumespresso.projwfc",
                ),
            }
        )


class BandsResourceSettingsPanel(
    PluginResourceSettingsPanel[BandsResourceSettingsModel],
):
    """Panel for configuring the band structure plugin."""



from importlib.metadata import distributions


def print_error(entry_point, e):
    print(f"\033[91mFailed to load plugin entry point {entry_point.name}.\033[0m")
    print(
        "\033[93mThis may be due to compatibility issues with the current QEApp version.\033[0m"
    )
    print("\033[93mPlease contact the plugin author for further assistance.\033[0m")
    print(
        "\033[93mThus, the plugin will not be available. However, you can still use the rest of the app.\033[0m"
    )
    print(f"\033[91mError message: {e}\033[0m\n")


# load entry points
def get_entries(entry_point_name="aiidalab_qe.properties"):
    from importlib_metadata import entry_points

    entries = {}
    for entry_point in entry_points(group=entry_point_name):
        try:
            # Attempt to load the entry point
            if entry_point.name in entries:
                continue
            loaded_entry_point = entry_point.load()
            entries[entry_point.name] = loaded_entry_point
        except Exception as e:
            print_error(entry_point, e)

    return entries


def get_entry_items(entry_point_name, item_name="outline"):
    entries = get_entries(entry_point_name)
    return {
        name: entry_point.get(item_name)
        for name, entry_point in entries.items()
        if entry_point.get(item_name, False)
    }


def get_entry_points_for_package(
    package_name: str, group: str = "aiidalab_qe.properties"
):
    """Find the entry points for the specified package"""
    entry_points_list = []

    dist = next(
        (d for d in distributions() if d.metadata["Name"] == package_name), None
    )
    if not dist:
        raise ValueError(f"Package '{package_name}' not found.")
    # Retrieve all entry points associated with this distribution
    if dist.entry_points:
        for ep in dist.entry_points:
            if ep.group == group:
                entry_points_list.append(ep)
    else:
        print(f"No entry points found for package '{package_name}'.")
    return entry_points_list


def test_plugin_functionality(plugin_name):
    """Test the functionality of the plugin.
    1) loading all entry points.
    2) check if the plugin use correct QEAppComputationalResourcesWidget
    """
    from aiidalab_qe.common.widgets import QEAppComputationalResourcesWidget

    try:
        eps = get_entry_points_for_package(plugin_name)
        # check if we can load all entry points
        for ep in eps:
            loaded_ep = ep.load()
            # check if the code uses the correct widget
            for name, code in loaded_ep.get("code", {}).items():
                if not isinstance(code, QEAppComputationalResourcesWidget):
                    return (
                        False,
                        f"\nPlugin {plugin_name} code {name} must use QEAppComputationalResourcesWidget class",
                    )
    except Exception as e:
        return False, f"Failed to get entry points for package {plugin_name}: {e}"
    return True, ""



import base64
import pathlib
import tempfile
from threading import Thread

import ipywidgets as ipw


class DownloadDataWidget(ipw.VBox):
    def __init__(self, workchain_node):
        #
        self.download_archive_button = ipw.Button(
            description="Download AiiDA archive.aiida data",
            icon="download",
            button_style="primary",
            disabled=False,
            tooltip="Download the AiiDA archive of the simulation, ready to be shared or imported into another AiiDA profile",
            layout=ipw.Layout(width="100%"),
        )
        self.download_archive_button.on_click(self._download_data_thread)

        self.download_raw_button = ipw.Button(
            description="Download AiiDA raw data (zip format)",
            icon="download",
            button_style="primary",
            disabled=False,
            tooltip="Download the raw data of the simulation, organized in intuitive directory paths.",
            layout=ipw.Layout(width="100%"),
        )
        try:
            # check that we can import the ProcessDumper (not implemented in old AiiDA versions)
            # pre-commit: allow any unused imports in the next line
            from aiida.tools.dumping.processes import ProcessDumper  # noqa: F401

            self.download_raw_button.on_click(self._download_data_thread)
            self.dumper_is_available = True
        except Exception:
            self.dumper_is_available = False

        self.download_raw_button.disabled = not self.dumper_is_available

        self.node = workchain_node

        self._downloading_message = ipw.HTML()

        children = []

        if not self.dumper_is_available:
            children.append(
                ipw.HTML("""
                    <p style="color:red; line-height: 140%;">
                        The raw data download is not available because the AiiDA
                        version is too old.
                    </p>
                """),
            )

        children.extend(
            [
                ipw.HBox(children=[self.download_raw_button]),
                ipw.HBox(children=[self.download_archive_button]),
                self._downloading_message,
            ]
        )

        super().__init__(
            children=children,
        )

    def _download_data_thread(self, button_instance):
        thread = Thread(target=lambda: self._download_data(button_instance))
        thread.start()

    def _download_data(self, button_instance):
        """
        This method handles the download process when a download button is clicked.
        It updates the button's description to indicate that the download is in progress,
        determines whether to download the archive or raw data based on the button's description,
        generates the appropriate bitstream from the specified node, initiates the download
        with a filename based on the node's primary key, and then resets the button description
        to its original state.

        Args:
            button_instance (ipywidgets.Button): The button instance that was clicked.
        """
        if "archive" in button_instance.description:
            what = "archive"
            filename = f"export_qeapp_calculation_pk_{self.node.pk}.aiida"
        else:
            what = "raw"
            filename = f"export_{self.node.pk}_raw.zip"

        self._disable_buttons()
        self._show_downloading_message(what)
        data = self.produce_bitestream(self.node, what=what)
        self._download(payload=data, filename=filename)
        del data
        self._hide_downloading_message()
        self._enable_buttons()

    def _show_downloading_message(self, what):
        self._downloading_message.value = f"""
            <div style="display: flex; align-items: center; margin: 6px 0;">
                Creating {what} data to download
                <i class="fa fa-spinner fa-spin fa-2x fa-fw" style="margin-left: 4px;"></i>
            </div>
        """

    def _hide_downloading_message(self):
        self._downloading_message.value = ""

    def _disable_buttons(self):
        self.download_raw_button.disabled = True
        self.download_archive_button.disabled = True

    def _enable_buttons(self):
        if self.dumper_is_available:
            self.download_raw_button.disabled = False
        self.download_archive_button.disabled = False

    @staticmethod
    def _download(payload, filename):
        from IPython.display import Javascript, display

        javas = Javascript(
            f"""
            var link = document.createElement('a');
            link.href = 'data:application;base64,{payload}'
            link.download = '{filename}'
            document.body.appendChild(link);
            link.click();
            document.body.removeChild(link);
            """
        )
        display(javas)

    @staticmethod
    def produce_bitestream(node, what="archive"):
        """
        Produce a base64-encoded bitstream of the specified node data.

        Parameters:
        node (orm.Node): The AiiDA node to be processed.
        what (str): The type of data to produce. Options are "archive" or "raw".
                    Defaults to "archive".

        Returns:
        str: A base64-encoded string representing the requested data.

        Raises:
        KeyError: If the 'what' parameter is not "archive" or "raw".

        The function supports two modes:
        1. "archive": Creates an AiiDA archive of the node.
        2. "raw": Dumps the raw data of the process node into a zip file.

        NB: The function uses a temporary directory to store the data before converting it to a base64 string.
            Moreover, the node has to be reloaded because otherwise the SQLAlchemy will compleain on Db request
            not being in the same thread (the notebook session) as the original node.
        """
        from aiida import orm

        reloaded_node = orm.load_node(node.pk)
        with tempfile.TemporaryDirectory() as dirpath:
            if what == "archive":
                from aiida.tools.archive.create import create_archive

                path = pathlib.Path(dirpath) / "archive.aiida"
                create_archive(
                    entities=[reloaded_node],
                    filename=path,
                    call_calc_backward=False,
                    call_work_backward=False,
                    create_backward=False,
                )
                with open(path, "rb") as f:
                    zip_data = f.read()

                # Convert the ZIP data to base64 so it can be used as a payload in JavaScript
                bitestream = base64.b64encode(zip_data).decode()

            elif what == "raw":
                import shutil

                from aiida.tools.dumping.processes import ProcessDumper

                path = pathlib.Path(dirpath) / "raw_data"
                output_zip_path = pathlib.Path(dirpath) / "raw_data.zip"
                dumper = ProcessDumper()
                dumper.dump(process_node=reloaded_node, output_path=path)
                # writing files to a zipfile
                shutil.make_archive(pathlib.Path(dirpath) / "raw_data", "zip", path)

                with open(output_zip_path, "rb") as f:
                    raw_data = f.read()

                # Convert the raw_data to base64 so it can be used as a payload in JavaScript
                bitestream = base64.b64encode(raw_data).decode()

            else:
                raise KeyError("You should ask for `archive` or `raw` only!")

            return bitestream



from __future__ import annotations

import shutil
from pathlib import Path
from tempfile import TemporaryDirectory

import ipywidgets as ipw
import traitlets as tl
from filelock import FileLock, Timeout
from IPython.display import HTML, display

from aiida import orm
from aiida.common import LinkType

from .download_data import DownloadDataWidget


class WorkChainOutputs(ipw.VBox):
    _busy = tl.Bool(read_only=True)

    def __init__(self, node, export_dir=None, **kwargs):
        if export_dir is None:
            export_dir = Path.cwd().joinpath("exports")
        self.export_dir = export_dir

        if node.process_label != "QeAppWorkChain":
            raise KeyError(str(node.node_type))

        self.node = node

        self._create_archive_indicator = ipw.HTML(
            """<button disabled>
                <i class="fa fa-spinner fa-spin" aria-hidden="true"></i>
                    Creating archive...
            </button>"""
        )
        self._download_button_widget = DownloadDataWidget(workchain_node=self.node)

        super().__init__(
            children=[self._download_button_widget],
            **kwargs,
        )

    @tl.default("_busy")
    def _default_busy(self):
        return False

    def _download_archive(self, _):
        fn_archive = self.export_dir.joinpath(str(self.node.uuid)).with_suffix(".zip")
        fn_lockfile = fn_archive.with_suffix(".lock")

        try:
            self.set_trait("_busy", True)
            # Create exports archive directory.
            fn_archive.parent.mkdir(parents=True, exist_ok=True)
            # Try to obtain lock for creating archive...
            with FileLock(fn_lockfile, timeout=0):
                # Check whether archive file already exists.
                if not fn_archive.is_file():
                    # Create archive file.
                    with TemporaryDirectory() as tmpdir:
                        self._prepare_calcjob_io(self.node, Path(tmpdir))
                        shutil.make_archive(fn_archive.with_suffix(""), "zip", tmpdir)
            Path(fn_lockfile).unlink()  # Delete lock file.
        except Timeout:
            # Failed to obtain lock, presuming some other process is working on it.
            with FileLock(fn_lockfile, timeout=20):
                assert fn_archive.is_file()
        finally:
            self.set_trait("_busy", False)

        link_id = f"dl_{self.node.uuid}"

        display(
            HTML(
                f"""
        <html>
        <body>
        <a
            id="{link_id}"
            href="{fn_archive.relative_to(Path.cwd())}"
            download="{fn_archive.stem}"
        ></a>
        <script>
        (function download() {{document.getElementById("{id}").click();
        }})()
        </script>

        </body>
        </html>
        """
            )
        )

    @classmethod
    def _prepare_calcjob_io(cls, node: orm.WorkChainNode, root_folder: Path):
        """Prepare the calculation job input and output files.

        :param node: QeAppWorkChain node.
        """
        counter = 1

        for link1 in node.get_outgoing(link_type=LinkType.CALL_WORK):
            wc_node = link1.node
            for link2 in wc_node.get_outgoing(link_type=LinkType.CALL_WORK):
                base_node = link2.node
                base_label = (
                    f"iter{link2.link_label[-1]}"
                    if link1.link_label == "relax"
                    else link2.link_label
                )
                for link3 in base_node.get_outgoing(
                    link_type=LinkType.CALL_CALC, link_label_filter="iteration_%"
                ):
                    counter_str = f"0{counter}" if counter < 10 else str(counter)
                    pw_label = f"pw{link3.link_label[-1]}"

                    fdname = f"{counter_str}-{link1.link_label}-{base_label}-{pw_label}"

                    folder_path = root_folder / fdname

                    cls._write_calcjob_io(link3.node, folder_path)

                    counter += 1

    @staticmethod
    def _write_calcjob_io(calcjob: orm.CalcJobNode, folder: Path) -> None:
        """Write the ``calcjob`` in and output files to ``folder``.

        :param calcjob: calculation job node for which to write the IO files.
        :param folder: folder to which to write the IO files.
        """
        folder.mkdir(exist_ok=True)
        input_filepath = folder / "aiida.in"

        with calcjob.open(calcjob.get_option("input_filename"), "r") as ihandle:
            with input_filepath.open("w") as ohandle:
                ohandle.write(ihandle.read())

        pseudo_folder = folder / "pseudo"
        pseudo_folder.mkdir(exist_ok=True)

        for _, pseudo in calcjob.inputs.pseudos.items():
            pseudo_path = pseudo_folder / pseudo.filename

            with pseudo_path.open("w") as handle:
                handle.write(pseudo.get_content())

        retrieved = calcjob.outputs.retrieved

        for filename in retrieved.list_object_names():
            out_filepath = folder / filename
            with out_filepath.open("w") as handle:
                handle.write(retrieved.get_object_content(filename))



from .model import SmearingConfigurationSettingsModel
from .smearing import SmearingConfigurationSettingsPanel

__all__ = [
    "SmearingConfigurationSettingsModel",
    "SmearingConfigurationSettingsPanel",
]



from __future__ import annotations

from copy import deepcopy

import numpy as np
import traitlets as tl
from pymatgen.core.periodic_table import Element

from aiida_quantumespresso.data.hubbard_structure import HubbardStructureData
from aiidalab_qe.common.mixins import HasInputStructure

from ..subsettings import AdvancedCalculationSubSettingsModel


class HubbardConfigurationSettingsModel(
    AdvancedCalculationSubSettingsModel,
    HasInputStructure,
):
    identifier = "hubbard"

    dependencies = [
        "input_structure",
    ]

    is_active = tl.Bool(False)
    has_eigenvalues = tl.Bool(False)
    parameters = tl.Dict(
        key_trait=tl.Unicode(),  # kind name
        value_trait=tl.Float(),  # U value
        default_value={},
    )
    eigenvalue_options = tl.List(
        trait=tl.Unicode(),
        default_value=["-1", "0", "1"],
    )
    eigenvalues = tl.List(
        trait=tl.List(),  # [[[[state, spin, kind name, eigenvalue] # state] # spin] # kind name]
        default_value=[],
    )

    applicable_kind_names = []
    orbital_labels = []

    def update(self, specific=""):  # noqa: ARG002
        if not self.has_structure:
            self.applicable_kind_names = []
            self.orbital_labels = []
            self._defaults |= {
                "parameters": {},
                "eigenvalues": [],
            }
        else:
            self.orbital_labels = self._define_orbital_labels()
            if isinstance(self.input_structure, HubbardStructureData):
                self._defaults["parameters"] = (
                    self.get_parameters_from_hubbard_structure()
                )
                self.is_active = True
            else:
                self._defaults["parameters"] = self._define_default_parameters()
            self.applicable_kind_names = self._define_applicable_kind_names()
            self._defaults["eigenvalues"] = self._define_default_eigenvalues()
        with self.hold_trait_notifications():
            self.parameters = self._get_default_parameters()
            self.eigenvalues = self._get_default_eigenvalues()
            self.needs_eigenvalues_widget = len(self.applicable_kind_names) > 0

    def get_active_eigenvalues(self):
        if not (
            active_eigenvalues := [
                orbital_eigenvalue
                for element_eigenvalues in self.eigenvalues
                for spin_row in element_eigenvalues
                for orbital_eigenvalue in spin_row
                if orbital_eigenvalue[-1] != -1
            ]
        ):
            return []
        eigenvalues_array = np.array(active_eigenvalues, dtype=object)
        new_shape = (int(np.prod(eigenvalues_array.shape[:-1])), 4)
        return [
            tuple(eigenvalue)
            for eigenvalue in eigenvalues_array.reshape(new_shape).tolist()
        ]

    def set_active_eigenvalues(self, eigenvalues: list):
        eigenvalues_array = np.array(eigenvalues, dtype=object)
        num_states = len(set(eigenvalues_array[:, 0]))
        num_spins = len(set(eigenvalues_array[:, 1]))
        num_kinds = len(set(eigenvalues_array[:, 2]))
        new_shape = (num_kinds, num_spins, num_states, 4)
        self.eigenvalues = eigenvalues_array.reshape(new_shape).tolist()
        self.has_eigenvalues = True

    def get_parameters_from_hubbard_structure(self):
        hubbard_parameters = self.input_structure.hubbard.dict()["parameters"]
        sites = self.input_structure.sites
        return {
            f"{sites[hp['atom_index']].kind_name} - {hp['atom_manifold']}": hp["value"]
            for hp in hubbard_parameters
        }

    def reset(self):
        with self.hold_trait_notifications():
            self.is_active = False
            self.has_eigenvalues = False
            self.parameters = self._get_default_parameters()
            self.eigenvalues = self._get_default_eigenvalues()

    def _define_orbital_labels(self):
        hubbard_manifold_list = [
            self._get_manifold(Element(kind.symbol))
            for kind in self.input_structure.kinds
        ]
        return [
            f"{kind_name} - {manifold}"
            for kind_name, manifold in zip(
                self.input_structure.get_kind_names(),
                hubbard_manifold_list,
            )
        ]

    def _define_default_parameters(self):
        return {label: 0.0 for label in self.orbital_labels}

    def _define_applicable_kind_names(self):
        applicable_kind_names = []
        for kind in self.input_structure.kinds:
            element = Element(kind.symbol)
            if (
                element.is_transition_metal
                or element.is_lanthanoid
                or element.is_actinoid
            ):
                num_states = 5 if element.is_transition_metal else 7
                applicable_kind_names.append((kind.name, num_states))
        return applicable_kind_names

    def _define_default_eigenvalues(self):
        return [
            [
                [
                    [state + 1, spin, kind_name, -1]  # default eigenvalue
                    for state in range(num_states)
                ]
                for spin in [1, 2]  # spin up and down
            ]
            for kind_name, num_states in self.applicable_kind_names  # transition metals and lanthanoids
        ]

    def _get_default_parameters(self):
        return deepcopy(self._defaults["parameters"])

    def _get_default_eigenvalues(self):
        return deepcopy(self._defaults["eigenvalues"])

    def _get_manifold(self, element):
        valence = [
            orbital
            for orbital in element.electronic_structure.split(".")
            if "[" not in orbital
        ]
        orbital_shells = [shell[:2] for shell in valence]

        def is_condition_met(shell):
            return condition and condition in shell

        # Conditions for determining the Hubbard manifold
        # to be selected from the electronic structure
        conditions = {
            element.is_transition_metal: "d",
            element.is_lanthanoid or element.is_actinoid: "f",
            element.is_post_transition_metal
            or element.is_metalloid
            or element.is_halogen
            or element.is_chalcogen
            or element.symbol in ["C", "N", "P"]: "p",
            element.is_alkaline or element.is_alkali or element.is_noble_gas: "s",
            element.symbol in ["H", "He"]: "s",
        }

        condition = next(
            (shell for condition, shell in conditions.items() if condition), None
        )

        hubbard_manifold = next(
            (shell for shell in orbital_shells if is_condition_met(shell)), None
        )

        return hubbard_manifold



from copy import deepcopy

import traitlets as tl
from aiida_pseudo.groups.family import PseudoPotentialFamily

from aiida_quantumespresso.workflows.protocols.utils import (
    get_magnetization_parameters,
    get_starting_magnetization,
)
from aiidalab_qe.common.mixins import HasInputStructure
from aiidalab_qe.utils import fetch_pseudo_family_by_label

from ..subsettings import AdvancedCalculationSubSettingsModel


class MagnetizationConfigurationSettingsModel(
    AdvancedCalculationSubSettingsModel,
    HasInputStructure,
):
    identifier = "magnetization"

    dependencies = [
        "input_structure",
        "electronic_type",
        "spin_type",
        "pseudos.family",
    ]

    electronic_type = tl.Unicode()
    spin_type = tl.Unicode()
    family = tl.Unicode()

    type_options = tl.List(
        trait=tl.List(tl.Unicode()),
        default_value=[
            ["Initial magnetic moments", "starting_magnetization"],
            ["Total magnetization", "tot_magnetization"],
        ],
    )
    type = tl.Unicode("starting_magnetization")
    type_help = tl.Unicode("")
    total = tl.Float(0.0)
    moments = tl.Dict(
        key_trait=tl.Unicode(),  # kind name
        value_trait=tl.Float(),  # magnetic moment
        default_value={},
    )

    _TYPE_HELP_TEMPLATE = """
        <div style="line-height: 1.4; margin-bottom: 5px">
            {content}
        </div>
    """

    _DEFAULT_MOMENTS = get_magnetization_parameters()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._default_starting_magnetization = {}

    def update(self, specific=""):  # noqa: ARG002
        if self.spin_type == "none" or not self.has_structure:
            self._defaults["moments"] = {}
        else:
            self.update_type_help()
            self.update_default_starting_magnetization()
            self._update_default_moments()
        with self.hold_trait_notifications():
            self.moments = self._get_default_moments()

    def update_type_help(self):
        """Update the type field help text w.r.t the current model state."""
        if self.electronic_type == "insulator" or self.type == "tot_magnetization":
            self.type_help = self._TYPE_HELP_TEMPLATE.format(
                content="""
                    Constrain the desired total electronic magnetization (difference
                    between majority and minority spin charge).
                """
            )
        else:
            self.type_help = self._TYPE_HELP_TEMPLATE.format(
                content="""
                    If a nonzero ground-state magnetization is expected, you
                    <strong>must</strong> assign a nonzero value to at least one atomic
                    type (the app already provides tentative initial values).
                    <br>
                    To simulate an antiferromagnetic state, first, if you have not
                    done so already, please use the atom tag editor <b>(Select
                    structure -> Edit structure -> Edit atom tags)</b> to mark atoms of
                    the species of interest as distinct by assigning each a different
                    integer tag. Once tagged, assign each an initial magnetic moments
                    of opposite sign.
                """
            )

    def update_default_starting_magnetization(self):
        """Update the default starting magnetization based on the structure and
        pseudopotential family.
        """
        if not self.has_structure:
            # TODO this guard shouldn't be here! It IS here only because in the present
            # implementation, an update is called on app start. This breaks lazy loading
            # and should be carefully checked!
            return
        family = fetch_pseudo_family_by_label(self.family)
        initial_guess = get_starting_magnetization(self.input_structure, family)
        self._default_starting_magnetization = initial_guess

    def reset(self):
        with self.hold_trait_notifications():
            self.type = self.traits()["type"].default_value
            self.total = self.traits()["total"].default_value
            self.moments = self._get_default_moments()

    def _update_default_moments(self):
        family = fetch_pseudo_family_by_label(self.family)
        self._defaults["moments"] = {
            kind.name: self._to_moment(symbol=kind.symbol, family=family)
            for kind in self.input_structure.kinds
        }

    def _to_moment(self, symbol: str, family: PseudoPotentialFamily) -> float:
        """Convert the default magnetization to an initial magnetic moment."""
        magnetization = (
            self._default_starting_magnetization.get(symbol, 0.1)
            if self._DEFAULT_MOMENTS.get(symbol, {}).get("magmom")
            else 0.1
        )
        return round(magnetization * family.get_pseudo(symbol).z_valence, 3)

    def _get_default_moments(self):
        return deepcopy(self._defaults.get("moments", {}))



import typing as t

import ipywidgets as ipw

from aiidalab_qe.common.mixins import HasProcess
from aiidalab_qe.common.mvc import Model
from aiidalab_qe.common.widgets import LoadingWidget


class ResultsComponentModel(Model, HasProcess):
    identifier = "results"


RCM = t.TypeVar("RCM", bound=ResultsComponentModel)


class ResultsComponent(ipw.VBox, t.Generic[RCM]):
    def __init__(self, model: RCM, **kwargs):
        self.loading_message = LoadingWidget(f"Loading {model.identifier}")

        super().__init__(
            children=[self.loading_message],
            **kwargs,
        )

        self._model = model
        self._model.observe(
            self._on_process_change,
            "process_uuid",
        )
        self._model.observe(
            self._on_monitor_counter_change,
            "monitor_counter",
        )

        self.rendered = False

    def render(self):
        if self.rendered:
            return
        self._render()
        self.rendered = True
        self._post_render()

    def _on_process_change(self, _):
        pass

    def _on_monitor_counter_change(self, _):
        pass

    def _render(self):
        raise NotImplementedError

    def _post_render(self):
        pass



import math

import numpy as np

from aiida import orm
from aiida.common import AttributeDict
from aiida.engine import WorkChain
from aiida.plugins import DataFactory, WorkflowFactory

GAMMA = "\u0393"

PwBandsWorkChain = WorkflowFactory("quantumespresso.pw.bands")
ProjwfcBandsWorkChain = WorkflowFactory("wannier90_workflows.projwfcbands")
KpointsData = DataFactory("core.array.kpoints")


def points_per_branch(vector_a, vector_b, reciprocal_cell, bands_kpoints_distance):
    """function to calculate the number of points per branch depending on the kpoints_distance and the reciprocal cell"""
    scaled_vector_a = np.array(vector_a)
    scaled_vector_b = np.array(vector_b)
    reciprocal_vector_a = scaled_vector_a.dot(reciprocal_cell)
    reciprocal_vector_b = scaled_vector_b.dot(reciprocal_cell)
    distance = np.linalg.norm(reciprocal_vector_a - reciprocal_vector_b)
    return max(
        2, int(np.round(distance / bands_kpoints_distance))
    )  # at least two points for each segment, including both endpoints explicitly


def calculate_bands_kpoints_distance(kpoints_distance):
    """function to calculate the bands_kpoints_distance depending on the kpoints_distance"""
    if kpoints_distance >= 0.5:
        return 0.1
    elif 0.15 < kpoints_distance < 0.5:
        return 0.025
    else:
        return 0.015


def generate_kpath_1d(structure, kpoints_distance):
    """Return a kpoints object for one dimensional systems (from Gamma to X)
    The number of kpoints is calculated based on the kpoints_distance (as in the PwBandsWorkChain protocol)
    """
    kpoints = KpointsData()
    kpoints.set_cell_from_structure(structure)
    reciprocal_cell = kpoints.reciprocal_cell
    bands_kpoints_distance = calculate_bands_kpoints_distance(kpoints_distance)

    # Number of points per branch
    num_points_per_branch = points_per_branch(
        [0.0, 0.0, 0.0],
        [0.5, 0.0, 0.0],
        reciprocal_cell,
        bands_kpoints_distance,
    )
    # Generate the kpoints
    points = np.linspace(
        start=[0.0, 0.0, 0.0],
        stop=[0.5, 0.0, 0.0],
        endpoint=True,
        num=num_points_per_branch,
    )
    kpoints.set_kpoints(points.tolist())
    kpoints.labels = [[0, GAMMA], [len(points) - 1, "X"]]
    return kpoints


def generate_kpath_2d(structure, kpoints_distance, kpath_2d):
    """Return a kpoints object for two dimensional systems based on the selected 2D symmetry path
    The number of kpoints is calculated based on the kpoints_distance (as in the PwBandsWorkChain protocol)
    The 2D symmetry paths are defined as in The Journal of Physical Chemistry Letters 2022 13 (50), 11581-11594 (https://pubs.acs.org/doi/10.1021/acs.jpclett.2c02972)
    """
    kpoints = KpointsData()
    kpoints.set_cell_from_structure(structure)
    reciprocal_cell = kpoints.reciprocal_cell
    bands_kpoints_distance = calculate_bands_kpoints_distance(kpoints_distance)

    # dictionary with the 2D symmetry paths
    selected_paths = {
        "hexagonal": {
            "path": [
                [0.0, 0.0, 0.0],
                [0.5, 0.0, 0.0],
                [0.33333, 0.33333, 0.0],
                [1.0, 0.0, 0.0],
            ],
            "labels": [GAMMA, "M", "K", GAMMA],
        },
        "square": {
            "path": [
                [0.0, 0.0, 0.0],
                [0.5, 0.0, 0.0],
                [0.5, 0.5, 0.0],
                [1.0, 0.0, 0.0],
            ],
            "labels": [GAMMA, "X", "M", GAMMA],
        },
        "rectangular": {
            "path": [
                [0.0, 0.0, 0.0],
                [0.5, 0.0, 0.0],
                [0.5, 0.5, 0.0],
                [0.0, 0.5, 0.0],
                [1.0, 0.0, 0.0],
            ],
            "labels": [GAMMA, "X", "S", "Y", GAMMA],
        },
    }
    # if the selected path is centered_rectangular or oblique, the path is calculated based on the reciprocal cell
    if kpath_2d in ["centered_rectangular", "oblique"]:
        a1 = reciprocal_cell[0]
        a2 = reciprocal_cell[1]
        norm_a1 = np.linalg.norm(a1)
        norm_a2 = np.linalg.norm(a2)
        cos_gamma = (
            a1.dot(a2) / (norm_a1 * norm_a2)
        )  # Angle between a1 and a2 # like in https://pubs.acs.org/doi/10.1021/acs.jpclett.2c02972
        gamma = np.arccos(cos_gamma)
        eta = (1 - (norm_a1 / norm_a2) * cos_gamma) / (2 * np.power(np.sin(gamma), 2))
        nu = 0.5 - (eta * norm_a2 * cos_gamma) / norm_a1
        selected_paths["centered_rectangular"] = {
            "path": [
                [0.0, 0.0, 0.0],
                [0.5, 0.0, 0.0],
                [1 - eta, nu, 0],
                [0.5, 0.5, 0.0],
                [eta, 1 - nu, 0.0],
                [1.0, 0.0, 0.0],
            ],
            "labels": [GAMMA, "X", "H_1", "C", "H", GAMMA],
        }
        selected_paths["oblique"] = {
            "path": [
                [0.0, 0.0, 0.0],
                [0.5, 0.0, 0.0],
                [1 - eta, nu, 0],
                [0.5, 0.5, 0.0],
                [eta, 1 - nu, 0.0],
                [0.0, 0.5, 0.0],
                [1.0, 0.0, 0.0],
            ],
            "labels": [GAMMA, "X", "H_1", "C", "H", "Y", GAMMA],
        }
    path = selected_paths[kpath_2d]["path"]
    labels = selected_paths[kpath_2d]["labels"]
    branches = zip(path[:-1], path[1:])

    all_kpoints = []  # List to hold all k-points
    label_map = []  # List to hold labels and their corresponding k-point indices

    # Calculate the number of points per branch and generate the kpoints
    index_offset = 0  # Start index for each segment
    for (start, end), label_start, _ in zip(branches, labels[:-1], labels[1:]):
        num_points_per_branch = points_per_branch(
            start, end, reciprocal_cell, bands_kpoints_distance
        )
        # Exclude endpoint except for the last segment to prevent duplication
        points = np.linspace(start, end, num=num_points_per_branch, endpoint=False)
        all_kpoints.extend(points)
        label_map.append(
            (index_offset, label_start)
        )  # Label for the start of the segment
        index_offset += len(points)

    # Include the last point and its label
    all_kpoints.append(path[-1])
    label_map.append((index_offset, labels[-1]))  # Label for the last point

    # Set the kpoints and their labels in KpointsData
    kpoints.set_kpoints(all_kpoints)
    kpoints.labels = label_map

    return kpoints


def determine_symmetry_path(structure):
    # Tolerance for checking equality
    cell_lengths = structure.cell_lengths
    cell_angles = structure.cell_angles
    tolerance = 1e-3

    # Define symmetry conditions and their corresponding types in a dictionary
    symmetry_conditions = {
        (
            math.isclose(cell_lengths[0], cell_lengths[1], abs_tol=tolerance)
            and math.isclose(cell_angles[2], 120.0, abs_tol=tolerance)
        ): "hexagonal",
        (
            math.isclose(cell_lengths[0], cell_lengths[1], abs_tol=tolerance)
            and math.isclose(cell_angles[2], 90.0, abs_tol=tolerance)
        ): "square",
        (
            not math.isclose(cell_lengths[0], cell_lengths[1], abs_tol=tolerance)
            and math.isclose(cell_angles[2], 90.0, abs_tol=tolerance)
        ): "rectangular",
        (
            math.isclose(
                cell_lengths[1] * math.cos(math.radians(cell_angles[2])),
                cell_lengths[0] / 2,
                abs_tol=tolerance,
            )
        ): "rectangular_centered",
        (
            not math.isclose(cell_lengths[0], cell_lengths[1], abs_tol=tolerance)
            and not math.isclose(cell_angles[2], 90.0, abs_tol=tolerance)
        ): "oblique",
    }

    # Check for symmetry type based on conditions
    for condition, symmetry_type in symmetry_conditions.items():
        if condition:
            return symmetry_type

    raise ValueError("Invalid symmetry type")


class BandsWorkChain(WorkChain):
    "Workchain to compute the electronic band structure"

    label = "bands"

    @classmethod
    def define(cls, spec):
        super().define(spec)

        spec.input("structure", valid_type=orm.StructureData)
        spec.expose_inputs(
            PwBandsWorkChain,
            namespace="bands",
            exclude=["structure", "relax"],
            namespace_options={
                "required": False,
                "populate_defaults": False,
                "help": "Inputs for the `PwBandsWorkChain`, simulation mode normal.",
            },
        )
        spec.expose_inputs(
            ProjwfcBandsWorkChain,
            namespace="bands_projwfc",
            exclude=["structure", "relax"],
            namespace_options={
                "required": False,
                "populate_defaults": False,
                "help": "Inputs for the `ProjwfcBandsWorkChain`, simulation mode fat_bands.",
            },
        )

        spec.expose_outputs(
            PwBandsWorkChain,
            namespace="bands",
            namespace_options={
                "required": False,
                "help": "Outputs of the `PwBandsWorkChain`.",
            },
        )
        spec.expose_outputs(
            ProjwfcBandsWorkChain,
            namespace="bands_projwfc",
            namespace_options={
                "required": False,
                "help": "Outputs of the `PwBandsWorkChain`.",
            },
        )

        spec.outline(cls.setup, cls.run_bands, cls.results)

        spec.exit_code(
            400, "ERROR_WORKCHAIN_FAILED", message="The workchain bands failed."
        )

    @classmethod
    def get_builder_from_protocol(
        cls,
        pw_code,
        projwfc_code,
        structure,
        simulation_mode="normal",
        protocol=None,
        overrides=None,
        **kwargs,
    ):
        """Return a BandsWorkChain builder prepopulated with inputs following the specified protocol

        :param structure: the ``StructureData`` instance to use.
        :param pw_code: the ``Code`` instance configured for the ``quantumespresso.pw`` plugin.
        :param protocol: protocol to use, if not specified, the default will be used.
        :param projwfc_code: the ``Code`` instance configured for the ``quantumespresso.projwfc`` plugin.
        :param simulation_mode: hat type of simulation to run normal band or fat bands.

        """

        builder = cls.get_builder()

        if simulation_mode == "normal":
            builder_bands = PwBandsWorkChain.get_builder_from_protocol(
                pw_code, structure, protocol, overrides=overrides, **kwargs
            )
            builder.pop("bands_projwfc", None)
            builder_bands.pop("relax", None)
            builder_bands.pop("structure", None)
            builder.bands = builder_bands

        elif simulation_mode == "fat_bands":
            builder_bands_projwfc = ProjwfcBandsWorkChain.get_builder_from_protocol(
                pw_code,
                projwfc_code,
                structure,
                protocol=protocol,
                overrides=overrides,
                **kwargs,
            )
            builder.pop("bands", None)
            builder_bands_projwfc.pop("relax", None)
            builder_bands_projwfc.pop("structure", None)
            builder.bands_projwfc = builder_bands_projwfc

        else:
            raise ValueError(f"Unknown simulation_mode: {simulation_mode}")

        # Handle periodic boundary conditions (PBC)
        if structure.pbc != (True, True, True):
            kpoints_distance = overrides.get("scf", {}).get(
                "kpoints_distance",
                builder.get("bands", {}).get("scf", {}).get("kpoints_distance"),
            )
            kpoints = None
            if structure.pbc == (True, False, False):
                kpoints = generate_kpath_1d(structure, kpoints_distance)
            elif structure.pbc == (True, True, False):
                kpoints = generate_kpath_2d(
                    structure=structure,
                    kpoints_distance=kpoints_distance,
                    kpath_2d=determine_symmetry_path(structure),
                )

            if simulation_mode == "normal":
                builder.bands.pop("bands_kpoints_distance", None)
                builder.bands.update({"bands_kpoints": kpoints})
            elif simulation_mode == "fat_bands":
                builder.bands_projwfc.pop("bands_kpoints_distance", None)
                builder.bands_projwfc.update({"bands_kpoints": kpoints})

        builder.structure = structure
        return builder

    def setup(self):
        """Define the current workchain"""
        self.ctx.current_structure = self.inputs.structure
        if "bands" in self.inputs:
            self.ctx.key = "bands"
            self.ctx.workchain = PwBandsWorkChain
        elif "bands_projwfc" in self.inputs:
            self.ctx.key = "bands_projwfc"
            self.ctx.workchain = ProjwfcBandsWorkChain
        else:
            self.report("No bands workchain specified")
            return self.exit_codes.ERROR_WORKCHAIN_FAILED

    def run_bands(self):
        """Run the bands workchain"""
        inputs = AttributeDict(
            self.exposed_inputs(self.ctx.workchain, namespace=self.ctx.key)
        )
        inputs.metadata.call_link_label = self.ctx.key
        inputs.structure = self.ctx.current_structure
        future = self.submit(self.ctx.workchain, **inputs)
        self.report(f"submitting `WorkChain` <PK={future.pk}>")
        self.to_context(**{self.ctx.key: future})

    def results(self):
        """Attach the bands results"""
        workchain = self.ctx[self.ctx.key]

        if not workchain.is_finished_ok:
            self.report("Bands workchain failed")
            return self.exit_codes.ERROR_WORKCHAIN_FAILED
        else:
            self.out_many(
                self.exposed_outputs(
                    self.ctx[self.ctx.key], self.ctx.workchain, namespace=self.ctx.key
                )
            )
            self.report("Bands workchain completed successfully")



import ipywidgets as ipw

from aiidalab_qe.common.widgets import HBoxWithUnits

from ..subsettings import AdvancedConfigurationSubSettingsPanel
from .model import MagnetizationConfigurationSettingsModel


class MagnetizationConfigurationSettingsPanel(
    AdvancedConfigurationSubSettingsPanel[MagnetizationConfigurationSettingsModel],
):
    """Widget to set the type of magnetization used in the calculation:
    1) Total magnetization: Total majority spin charge - minority spin charge.
    2) Magnetic moments: Starting spin polarization on atomic type 'i' in a spin polarized (LSDA or noncollinear/spin-orbit) calculation.

    For Magnetic moments you can set each kind names defined in the StructureData (StructureData.get_kind_names())
    Usually these are the names of the elements in the StructureData
    (For example 'C' , 'N' , 'Fe' . However the StructureData can have defined kinds like 'Fe1' and 'Fe2')
    The widget generate a dictionary that can be used to set initial_magnetic_moments in the builder of PwBaseWorkChain

    Attributes:
        input_structure(StructureData): trait that contains the input_structure (confirmed structure from previous step)
    """

    def __init__(self, model: MagnetizationConfigurationSettingsModel, **kwargs):
        super().__init__(model, **kwargs)

        self._model.observe(
            self._on_input_structure_change,
            "input_structure",
        )
        self._model.observe(
            self._on_electronic_type_change,
            "electronic_type",
        )
        self._model.observe(
            self._on_spin_type_change,
            "spin_type",
        )
        self._model.observe(
            self._on_magnetization_type_change,
            "type",
        )
        self._model.observe(
            self._on_family_change,
            "family",
        )

    def render(self):
        if self.rendered:
            return

        self.header = ipw.HTML("<h2>Magnetization</h2>")

        self.unit = "µ<sub>B</sub>"

        self.insulator_help = ipw.HTML("""
            <div style="line-height: 1.4;">
                <b>Note:</b> defining the starting magnetic moments per atomic species
                is available only for metallic systems. To enable the feature, please
                set the <b>Electronic type</b> to <b>Metal</b> in the <b>Basic
                settings</b> tab.
            </div>
        """)

        self.magnetization_type_help = ipw.HTML()
        ipw.dlink(
            (self._model, "type_help"),
            (self.magnetization_type_help, "value"),
        )

        self.magnetization_type = ipw.ToggleButtons(
            style={"button_width": "initial"},
            layout=ipw.Layout(margin="0 0 5px"),
        )
        ipw.dlink(
            (self._model, "type_options"),
            (self.magnetization_type, "options"),
        )
        ipw.link(
            (self._model, "type"),
            (self.magnetization_type, "value"),
        )

        self.tot_magnetization = ipw.BoundedIntText(
            min=0,
            max=100,
            step=1,
            description="Total magnetization:",
            style={"description_width": "150px"},
        )
        ipw.link(
            (self._model, "total"),
            (self.tot_magnetization, "value"),
        )

        self.tot_magnetization_with_unit = HBoxWithUnits(
            self.tot_magnetization,
            self.unit,
        )

        self.kind_moment_widgets = ipw.VBox()

        self.container = ipw.VBox(
            children=[
                self.tot_magnetization_with_unit,
            ]
        )

        self.children = []

        self.rendered = True

        self.refresh(specific="widgets")

    def _on_input_structure_change(self, _):
        self.refresh(specific="structure")

    def _on_electronic_type_change(self, _):
        self._switch_widgets()
        self._model.update_type_help()

    def _on_spin_type_change(self, _):
        self.refresh(specific="spin")

    def _on_magnetization_type_change(self, _):
        self._toggle_widgets()
        self._model.update_type_help()

    def _on_family_change(self, _):
        self._model.update_default_starting_magnetization()

    def _update(self, specific=""):
        if self.updated:
            return
        self._show_loading()
        if not self._model.loaded_from_process or (specific and specific != "widgets"):
            self._model.update(specific)
        self._build_kinds_widget()
        self._switch_widgets()
        self._toggle_widgets()
        self.updated = True

    def _show_loading(self):
        if self.rendered:
            self.kind_moment_widgets.children = [self.loading_message]

    def _build_kinds_widget(self):
        if not self.rendered:
            return

        children = []

        kind_names = (
            self._model.input_structure.get_kind_names()
            if self._model.input_structure
            else []
        )

        for kind_name in kind_names:
            kind_moment_widget = ipw.BoundedFloatText(
                description=kind_name,
                min=-7,
                max=7,
                step=0.1,
                style={"description_width": "150px"},
            )
            link = ipw.link(
                (self._model, "moments"),
                (kind_moment_widget, "value"),
                [
                    lambda moments, kind_name=kind_name: moments.get(kind_name, 0.0),
                    lambda value, kind_name=kind_name: {
                        **self._model.moments,
                        kind_name: value,
                    },
                ],
            )
            self.links.append(link)
            children.append(HBoxWithUnits(kind_moment_widget, "µ<sub>B</sub>"))

        self.kind_moment_widgets.children = children

    def _switch_widgets(self):
        if not self.rendered:
            return
        if self._model.spin_type == "none":
            children = []
        else:
            children = [self.header]
            if self._model.electronic_type == "metal":
                children.extend(
                    [
                        self.magnetization_type,
                        self.magnetization_type_help,
                        self.container,
                    ]
                )
            else:
                children.extend(
                    [
                        self.insulator_help,
                        self.magnetization_type_help,
                        self.tot_magnetization_with_unit,
                    ],
                )
        self.children = children

    def _toggle_widgets(self):
        if self._model.spin_type == "none" or not self.rendered:
            return
        self.container.children = [
            self.tot_magnetization_with_unit
            if self._model.type == "tot_magnetization"
            else self.kind_moment_widgets
        ]



from .result import ElectronicStructureResultsModel, ElectronicStructureResultsPanel

electronic_structure = {
    "result": {
        "panel": ElectronicStructureResultsPanel,
        "model": ElectronicStructureResultsModel,
    },
}
